"""Message bus configuration loader.

The message bus event handler and client use basically the same configuration.
This code is re-used in both to load config values.
"""
import json
from collections import namedtuple

from ovos_utils.log import LOG
from ovos_config.config import Configuration

# mycroft-core had this duplicated with both names...
MessageBusConfig = MessageBusClientConf = namedtuple('MessageBusClientConf',
                                                     ['host', 'port', 'route',
                                                      'ssl'])


def load_message_bus_config(**overrides) -> MessageBusConfig:
    """
    Load the bits of device configuration needed to run the message bus.
    @param overrides: Optional config overrides
        host (str): messagebus host
        port (int): messagebus port
        route (str): messagebus route
        ssl (bool): enable SSL on websocket
    @return: MessageBusConfig with valid configuration
    """
    LOG.debug('Loading message bus configs')
    config = Configuration()

    try:
        config = config['websocket']
    except KeyError as ke:
        LOG.error(f'No websocket configs found ({ke})')
        raise ke
    overrides = overrides or {}
    config = config or {}
    mb_config = MessageBusConfig(
        host=overrides.get('host') or config.get('host') or "127.0.0.1",
        port=overrides.get('port') or config.get('port') or 8181,
        route=overrides.get('route') or config.get('route') or "/core",
        ssl=overrides.get('ssl') if 'ssl' in overrides else
        config.get('ssl') if 'ssl' in config else
        False
    )

    return mb_config


def load_gui_message_bus_config(**overrides):
    """
    Load the bits of device configuration needed to run the GUI bus.
    @param overrides: Optional config overrides
        host (str): GUI ebus host
        port (int): GUI bus port
        route (str): GUI bus route
        ssl (bool): enable SSL on websocket
    @return: MessageBusConfig with valid configuration
    """
    LOG.info('Loading GUI bus configs')
    config = Configuration()

    try:
        config = config['gui']
    except KeyError as ke:
        LOG.error(f'No gui configs found ({ke})')
        raise
    overrides = overrides or {}
    config = config or {}
    mb_config = MessageBusConfig(
        host=overrides.get('host') or config.get('host') or "127.0.0.1",
        port=overrides.get('port') or config.get('port') or 18181,
        route=overrides.get('route') or config.get('route') or "/",
        ssl=overrides.get('ssl') if 'ssl' in overrides else
        config.get('ssl') if 'ssl' in config else
        False
    )

    return mb_config


def client_from_config(subconf: str = 'core',
                       file_path: str = '/etc/mycroft/bus.conf'):
    """
    Load messagebus configuration from file.

    The config is a basic json file with a number of "sub configurations"

    Ex:
    {
      "core": {
        "route": "/core",
        "port": "8181"
      }
      "gui": {
        "route": "/gui",
        "port": "8811"
      }
    }

    Arguments:
        subconf:    configuration to choose from the file, defaults to "core"
                    if omitted.
        file_path:  path to the config file, defaults to /etc/mycroft/bus.conf
                    if omitted.
    Returns:
        MessageBusClient instance based on the selected config.
    """
    from ovos_bus_client.client import MessageBusClient

    with open(file_path) as f:
        conf = json.load(f)

    return MessageBusClient(**conf[subconf])



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from queue import Queue
from threading import Lock, Event
from uuid import uuid4
import time

try:
    from mycroft_bus_client.client.collector import MessageCollector as _MessageCollectorBase
except ImportError:
    # TODO - code in the wild does isinstance checks
    # this conditional subclassing should be removed ASAP, it is only here for the migration period
    # mycroft_bus_client is abandonware until further notice from MycroftAI

    class _MessageCollectorBase:
        pass


class MessageCollector(_MessageCollectorBase):
    """Collect multiple response.

    This class encapsulates the logic for collecting messages from
    multiple handlers returning the list of all answers.

    Argunments:
        bus: Bus to check for messages on
        message (Message): message to send
        min_timeout (int/float): Minimum time to wait for a response
        max_timeout (int/float): Maximum allowed time to wait for an answer
        direct_return_func (callable): Optional function for allowing an
            early return (not all registered handlers need to respond)
    """
    def __init__(self, bus, message,
                 min_timeout, max_timeout,
                 direct_return_func=None):
        self.lock = Lock()
        self.bus = bus
        self.min_timeout = min_timeout
        self.max_timeout = max_timeout
        self.direct_return_func = direct_return_func or (lambda msg: False)

        # Create an unique id for the collection
        self.collect_id = str(uuid4())
        self.handlers = {}
        self.responses = {}
        self.all_collected = Event()
        self.message = message
        self.message.context['__collect_id__'] = self.collect_id
        self._start_time = 0

        self.on_response_callback = None
        self.queue = Queue()

    def __iter__(self):
        return self

    def __next__(self):
        msg = Queue.get()
        if msg is not None:
            return msg
        else:
            raise StopIteration

    def on_response(self, callback_func):
        self.on_response_callback = callback_func

    def _register_handler(self, msg):
        """Handler for registration of collection handler.

        Args:
            msg: Message from handler.
        """
        handler_id = msg.data['handler']
        timeout = msg.data['timeout']
        with self.lock:
            if (msg.data['query'] == self.collect_id and
                    handler_id not in self.handlers):
                previous_timeout = self.handlers.get(handler_id, 0)
                self.handlers[handler_id] = previous_timeout + timeout

    def _receive_response(self, msg):
        """Handler for capturing final response from a handler.

        Args:
            msg: Message with collect handler's response.
        """
        with self.lock:
            if msg.data['query'] == self.collect_id:
                self.queue.put(msg)
                self.responses[msg.data['handler']] = msg
                self.handlers[msg.data['handler']] = 0  # Reset timeout
                # If all registered handlers have responded with an answer
                # or a VERY good answer has been found indicate end of wait.
                all_collected = len(self.responses) == len(self.handlers)
                if (all_collected or self.direct_return_func(msg)):
                    self.queue.put(None)
                    self.all_collected.set()

        if self.on_response_callback:
            self.on_response_callback(msg)

    def _setup_collection_handlers(self):
        """Create messages for handling and responses."""
        base_msg_type = self.message.msg_type
        self.bus.on(base_msg_type + '.handling', self._register_handler)
        self.bus.on(base_msg_type + '.response', self._receive_response)

    def _teardown_collection_handlers(self):
        """Remove all registered handlers for response collection."""
        base_msg_type = self.message.msg_type
        self.bus.remove(base_msg_type + '.handling', self._register_handler)
        self.bus.remove(base_msg_type + '.response', self._receive_response)

    def start(self):
        """Send collection request.

        Register handler to capture handlers trying to provide answer.
        """
        self._setup_collection_handlers()
        self.bus.emit(self.message)
        self.start_time = time.monotonic()

        time.sleep(self.min_timeout)

    def collect(self):
        """Emit message and wait for handlers to finish."""
        self.start()
        if len(self.handlers) == 0:
            # No handlers has registered to answer the query
            result = []
        else:
            result = self._wait_for_registered_handlers()

        self.shutdown()
        return result

    def wait(self):
        """Wait for timeout or for all handlers to respond."""
        self._wait_for_registered_handlers()

    def _wait_for_registered_handlers(self):
        """
        Wait until all handlers have sent a response or the timeout is reached.
        """
        # Reset the all_collected event if needed.
        # May be set if the first registered message replies immediately before
        # any other handlers has registered.
        # TODO: check early return criteria
        with self.lock:
            all_collected = len(self.responses) == len(self.handlers)
            if not all_collected:
                self.all_collected.clear()

        # Wait until all handlers have responded or timeout is reached
        time_waited = self.min_timeout
        remaining_timeout = max(self.handlers.values()) - time_waited
        while remaining_timeout > 0.0 and time_waited < self.max_timeout:
            if self.all_collected.wait(timeout=0.1):
                break

            time_waited += 0.1
            remaining_timeout = max(self.handlers.values()) - time_waited

        self.queue.put(None)
        return [self.responses[key] for key in self.responses]

    def shutdown(self):
        """Shutdown the object, stop waiting for further responses."""
        self._teardown_collection_handlers()
        if self.on_response_callback:
            self.on_response_callback = None



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import time
from datetime import timedelta
from functools import wraps
from os.path import abspath
from threading import Lock
from typing import List, Union, Optional

from ovos_utils.gui import is_gui_connected, is_gui_running
from ovos_utils.log import LOG, deprecated

from ovos_bus_client.message import Message
from ovos_bus_client.message import dig_for_message
from ovos_bus_client.util import get_mycroft_bus


def ensure_uri(s: str):
    """
    Interpret paths as file:// uri's.

    Args:
        s: string path to be checked

    Returns:
        if s is uri, s is returned otherwise file:// is prepended
    """
    if isinstance(s, str):
        if ':' not in s:
            return 'file://' + abspath(s)
        else:
            return s
    elif isinstance(s, (tuple, list)):  # Handle (mime, uri) arg
        if ':' not in s[0]:
            return 'file://' + abspath(s[0]), s[1]
        else:
            return s
    else:
        raise ValueError('Invalid track')


def _ensure_message_kwarg():
    """ensure message kwarg is present
    NOTE: this is meant for usage only in this module, it is not a generic decorator!
    """

    def message_injector(func):
        # this method ensures all skills messages are .forward from the utterance
        # that triggered the skill, this ensures proper routing and metadata in message.context
        @wraps(func)
        def call_function(*args, **kwargs):
            if not any([isinstance(a, Message) for a in args]):
                m = kwargs.get("source_message")
                if not m:
                    source_message = dig_for_message(max_records=50)
                    if source_message:
                        kwargs["source_message"] = source_message
                    else:
                        LOG.warning("source message could not be determined, message.context has been lost!")
                        kwargs["source_message"] = Message("")
            return func(*args, **kwargs)

        return call_function

    return message_injector


class ClassicAudioServiceInterface:
    """AudioService class for interacting with the classic mycroft audio subsystem

    DEPRECATED: only works in ovos-core <= 0.0.8

    it has been removed from ovos-audio with the move to ovos-media
    
    "mycroft.audio.XXX" has been replaced by "ovos.audio.XXX" namespace

    use OCPInterface instead

    Args:
        bus: OpenVoiceOS messagebus connection
    """

    @deprecated("removed from ovos-audio with the adoption of ovos-media service, "
                "use OCPInterface instead", "0.1.0")
    def __init__(self, bus=None):
        self.bus = bus or get_mycroft_bus()

    @_ensure_message_kwarg()
    def queue(self, tracks=None, source_message: Optional[Message] = None):
        """Queue up a track to playing playlist.

        Args:
            tracks: track uri or list of track uri's
                    Each track can be added as a tuple with (uri, mime)
                    to give a hint of the mime type to the system
            source_message: bus message that triggered this action
        """
        tracks = tracks or []
        if isinstance(tracks, (str, tuple)):
            tracks = [tracks]
        elif not isinstance(tracks, list):
            raise ValueError
        tracks = [ensure_uri(t) for t in tracks]
        self.bus.emit(source_message.forward('mycroft.audio.service.queue',
                                             {'tracks': tracks}))

    @_ensure_message_kwarg()
    def play(self, tracks=None, utterance=None, repeat=None, source_message: Optional[Message] = None):
        """Start playback.

        Args:
            tracks: track uri or list of track uri's
                    Each track can be added as a tuple with (uri, mime)
                    to give a hint of the mime type to the system
            utterance: forward utterance for further processing by the
                        audio service.
            repeat: if the playback should be looped
            source_message: bus message that triggered this action
        """
        repeat = repeat or False
        tracks = tracks or []
        utterance = utterance or ''
        if isinstance(tracks, (str, tuple)):
            tracks = [tracks]
        elif not isinstance(tracks, list):
            raise ValueError
        tracks = [ensure_uri(t) for t in tracks]
        self.bus.emit(source_message.forward('mycroft.audio.service.play',
                                             {'tracks': tracks,
                                              'utterance': utterance,
                                              'repeat': repeat}))

    @_ensure_message_kwarg()
    def stop(self, source_message: Optional[Message] = None):
        """Stop the track.
         Args:
            source_message: bus message that triggered this action"""
        self.bus.emit(source_message.forward('mycroft.audio.service.stop'))

    @_ensure_message_kwarg()
    def next(self, source_message: Optional[Message] = None):
        """Change to next track.
         Args:
            source_message: bus message that triggered this action"""
        self.bus.emit(source_message.forward('mycroft.audio.service.next'))

    @_ensure_message_kwarg()
    def prev(self, source_message: Optional[Message] = None):
        """Change to previous track.
         Args:
            source_message: bus message that triggered this action"""
        self.bus.emit(source_message.forward('mycroft.audio.service.prev'))

    @_ensure_message_kwarg()
    def pause(self, source_message: Optional[Message] = None):
        """Pause playback.
         Args:
            source_message: bus message that triggered this action"""
        self.bus.emit(source_message.forward('mycroft.audio.service.pause'))

    @_ensure_message_kwarg()
    def resume(self, source_message: Optional[Message] = None):
        """Resume paused playback.
         Args:
            source_message: bus message that triggered this action"""
        self.bus.emit(source_message.forward('mycroft.audio.service.resume'))

    @_ensure_message_kwarg()
    def get_track_length(self, source_message: Optional[Message] = None):
        """
        getting the duration of the audio in seconds
         Args:
            source_message: bus message that triggered this action
        """
        length = 0
        info = self.bus.wait_for_response(
            source_message.forward('mycroft.audio.service.get_track_length'),
            timeout=1)
        if info:
            length = info.data.get("length") or 0
        return length / 1000  # convert to seconds

    @_ensure_message_kwarg()
    def get_track_position(self, source_message: Optional[Message] = None):
        """
        get current position in seconds
         Args:
            source_message: bus message that triggered this action
        """
        pos = 0
        info = self.bus.wait_for_response(
            source_message.forward('mycroft.audio.service.get_track_position'),
            timeout=1)
        if info:
            pos = info.data.get("position") or 0
        return pos / 1000  # convert to seconds

    @_ensure_message_kwarg()
    def set_track_position(self, seconds, source_message: Optional[Message] = None):
        """Seek X seconds.

        Arguments:
            seconds (int): number of seconds to seek, if negative rewind
            source_message: bus message that triggered this action
        """
        self.bus.emit(source_message.forward('mycroft.audio.service.set_track_position',
                                             {"position": seconds * 1000}))  # convert to ms

    @_ensure_message_kwarg()
    def seek(self, seconds: Union[int, float, timedelta] = 1,
             source_message: Optional[Message] = None):
        """Seek X seconds.

        Args:
            seconds (int): number of seconds to seek, if negative rewind
            source_message: bus message that triggered this action
        """
        if isinstance(seconds, timedelta):
            seconds = seconds.total_seconds()
        if seconds < 0:
            self.seek_backward(abs(seconds), source_message=source_message)
        else:
            self.seek_forward(seconds, source_message=source_message)

    @_ensure_message_kwarg()
    def seek_forward(self, seconds: Union[int, float, timedelta] = 1,
                     source_message: Optional[Message] = None):
        """Skip ahead X seconds.

        Args:
            seconds (int): number of seconds to skip
            source_message: bus message that triggered this action
        """
        if isinstance(seconds, timedelta):
            seconds = seconds.total_seconds()
        self.bus.emit(source_message.forward('mycroft.audio.service.seek_forward',
                                             {"seconds": seconds}))

    @_ensure_message_kwarg()
    def seek_backward(self, seconds: Union[int, float, timedelta] = 1, source_message: Optional[Message] = None):
        """Rewind X seconds

         Args:
            seconds (int): number of seconds to rewind
            source_message: bus message that triggered this action
        """
        if isinstance(seconds, timedelta):
            seconds = seconds.total_seconds()
        self.bus.emit(source_message.forward('mycroft.audio.service.seek_backward',
                                             {"seconds": seconds}))

    @_ensure_message_kwarg()
    def track_info(self, source_message: Optional[Message] = None):
        """Request information of current playing track.
         Args:
            source_message: bus message that triggered this action

        Returns:
            Dict with track info.
        """
        info = self.bus.wait_for_response(
            source_message.forward('mycroft.audio.service.track_info'),
            reply_type='mycroft.audio.service.track_info_reply',
            timeout=1)
        return info.data if info else {}

    @_ensure_message_kwarg()
    def available_backends(self, source_message: Optional[Message] = None):
        """Return available audio backends.
         Args:
            source_message: bus message that triggered this action

        Returns:
            dict with backend names as keys
        """
        m = source_message.forward('mycroft.audio.service.list_backends')
        response = self.bus.wait_for_response(m)
        return response.data if response else {}

    @property
    def is_playing(self):
        """True if the audioservice is playing, else False."""
        return self.track_info() != {}


class OCPInterface:
    """bus api interface for OCP subsystem
    Args:
        bus: OpenVoiceOS messagebus connection
    """

    def __init__(self, bus=None):
        self.bus = bus or get_mycroft_bus()

    # OCP bus api
    @staticmethod
    def norm_tracks(tracks: list):
        try:
            from ovos_utils.ocp import Playlist, MediaEntry, PluginStream, dict2entry
        except ImportError as e:
            raise RuntimeError("This method requires ovos-utils ~=0.1") from e

        """ensures a list of tracks contains only MediaEntry or Playlist items"""
        assert isinstance(tracks, list)
        # support Playlist and MediaEntry objects in tracks
        for idx, track in enumerate(tracks):
            if isinstance(track, dict):
                tracks[idx] = dict2entry(track)
            if isinstance(track, PluginStream):
                # TODO - this method will be deprecated
                #  once all SEI parsers can handle the new objects
                #  this module can serialize them just fine,
                #  but we dont know who is listening
                tracks[idx] = track.as_media_entry
            elif isinstance(track, list) and not isinstance(track, Playlist):
                tracks[idx] = OCPInterface.norm_tracks(track)
            elif not isinstance(track, (Playlist, MediaEntry)):
                # TODO - support string uris
                # let it fail in next assert
                # log all bad entries before failing
                LOG.error(f"Bad track, invalid type: {track}")
        assert all(isinstance(t, (MediaEntry, Playlist, PluginStream)) for t in tracks)
        return tracks

    @_ensure_message_kwarg()
    def queue(self, tracks: list, source_message: Optional[Message] = None):
        """Queue up a track to OCP playing playlist.

        Args:
            tracks: track dict or list of track dicts (OCP result style)
            source_message: bus message that triggered this action
        """
        tracks = self.norm_tracks(tracks)
        self.bus.emit(source_message.forward('ovos.common_play.playlist.queue',
                                             {'tracks': tracks}))

    @_ensure_message_kwarg()
    def populate_search_results(self, tracks: list,
                                replace: bool = True,
                                sort_by_conf: bool = True,
                                source_message: Optional[Message] = None):
        """populate search results in OCP with new tracks
        Args:
            tracks: track dict or list of track dicts (OCP result style)
            replace: if False, extend existing search, if True replace current search results
            source_message: bus message that triggered this action
        """
        tracks = self.norm_tracks(tracks)
        self.bus.emit(source_message.forward('ovos.common_play.search.populate',
                                             {"playlist": [t.as_dict for t in tracks],
                                              "replace": replace, "sort_by_conf": sort_by_conf}))

    @_ensure_message_kwarg()
    def play(self, tracks: list, utterance=None, source_message: Optional[Message] = None):
        """Start playback.
        Args:
            tracks: track dict or list of track dicts (OCP result style)
            utterance: forward utterance for further processing by OCP
            source_message: bus message that triggered this action
        """
        tracks = self.norm_tracks(tracks)
        utterance = utterance or ''
        playlist = tracks
        disambiguation = []
        try:
            from ovos_utils.ocp import Playlist, MediaEntry, PluginStream, dict2entry
            if isinstance(tracks[0], Playlist):
                playlist = tracks[0]
                disambiguation = tracks
        except ImportError as e:
            LOG.warning("can't handle Playlist results properly, please update ovos-utils to >= 0.1.0")

        media = playlist[0]
        self.bus.emit(source_message.forward('ovos.common_play.play',
                                             {"media": media.as_dict,
                                              "playlist": [t.as_dict for t in playlist],
                                              "disambiguation": [t.as_dict for t in disambiguation],
                                              "utterance": utterance}))

    @_ensure_message_kwarg()
    def stop(self, source_message: Optional[Message] = None):
        """Stop the track.
         Args:
            source_message: bus message that triggered this action"""
        self.bus.emit(source_message.forward("ovos.common_play.stop"))

    @_ensure_message_kwarg()
    def next(self, source_message: Optional[Message] = None):
        """Change to next track.
         Args:
            source_message: bus message that triggered this action"""
        self.bus.emit(source_message.forward("ovos.common_play.next"))

    @_ensure_message_kwarg()
    def prev(self, source_message: Optional[Message] = None):
        """Change to previous track.
         Args:
            source_message: bus message that triggered this action"""
        self.bus.emit(source_message.forward("ovos.common_play.previous"))

    @_ensure_message_kwarg()
    def pause(self, source_message: Optional[Message] = None):
        """Pause playback.
         Args:
            source_message: bus message that triggered this action"""
        self.bus.emit(source_message.forward("ovos.common_play.pause"))

    @_ensure_message_kwarg()
    def resume(self, source_message: Optional[Message] = None):
        """Resume paused playback.
         Args:
            source_message: bus message that triggered this action"""
        self.bus.emit(source_message.forward("ovos.common_play.resume"))

    @_ensure_message_kwarg()
    def seek_forward(self, seconds=1, source_message: Optional[Message] = None):
        """Skip ahead X seconds.
        Args:
            seconds (int): number of seconds to skip
            source_message: bus message that triggered this action
        """
        if isinstance(seconds, timedelta):
            seconds = seconds.total_seconds()
        self.bus.emit(source_message.forward('ovos.common_play.seek',
                                             {"seconds": seconds}))

    @_ensure_message_kwarg()
    def seek_backward(self, seconds=1, source_message: Optional[Message] = None):
        """Rewind X seconds
         Args:
            seconds (int): number of seconds to rewind
            source_message: bus message that triggered this action
        """
        if isinstance(seconds, timedelta):
            seconds = seconds.total_seconds()
        self.bus.emit(source_message.forward('ovos.common_play.seek',
                                             {"seconds": seconds * -1}))

    @_ensure_message_kwarg()
    def get_track_length(self, source_message: Optional[Message] = None):
        """
        getting the duration of the audio in miliseconds
         Args:
            source_message: bus message that triggered this action
        """
        length = 0
        msg = source_message.forward('ovos.common_play.get_track_length')
        info = self.bus.wait_for_response(msg, timeout=1)
        if info:
            length = info.data.get("length", 0)
        return length

    @_ensure_message_kwarg()
    def get_track_position(self, source_message: Optional[Message] = None):
        """
        get current position in miliseconds
         Args:
            source_message: bus message that triggered this action
        """
        pos = 0
        msg = source_message.forward('ovos.common_play.get_track_position')
        info = self.bus.wait_for_response(msg, timeout=1)
        if info:
            pos = info.data.get("position", 0)
        return pos

    @_ensure_message_kwarg()
    def set_track_position(self, miliseconds, source_message: Optional[Message] = None):
        """Go to X position.
        Arguments:
            miliseconds (int): position to go to in miliseconds
            source_message: bus message that triggered this action
        """
        self.bus.emit(source_message.forward('ovos.common_play.set_track_position',
                                             {"position": miliseconds}))

    @_ensure_message_kwarg()
    def track_info(self, source_message: Optional[Message] = None):
        """Request information of current playing track.
         Args:
            source_message: bus message that triggered this action
        Returns:
            Dict with track info.
        """
        msg = source_message.forward('ovos.common_play.track_info')
        response = self.bus.wait_for_response(msg)
        return response.data if response else {}

    @_ensure_message_kwarg()
    def available_backends(self, source_message: Optional[Message] = None):
        """Return available audio backends.
         Args:
            source_message: bus message that triggered this action
        Returns:
            dict with backend names as keys
        """
        msg = source_message.forward('ovos.common_play.list_backends')
        response = self.bus.wait_for_response(msg)
        return response.data if response else {}


class OCPQuery:
    try:
        from ovos_utils.ocp import MediaType
        cast2audio = [
            MediaType.MUSIC,
            MediaType.PODCAST,
            MediaType.AUDIOBOOK,
            MediaType.RADIO,
            MediaType.RADIO_THEATRE,
            MediaType.VISUAL_STORY,
            MediaType.NEWS
        ]
    except ImportError as e:
        from enum import IntEnum

        class MediaType(IntEnum):
            GENERIC = 0  # nothing else matches

        cast2audio = None

    def __init__(self, query, bus, media_type=MediaType.GENERIC, config=None):
        if self.cast2audio is None:
            raise RuntimeError("This class requires ovos-utils ~=0.1")
        LOG.debug(f"Created {media_type.name} query: {query}")
        self.query = query
        self.media_type = media_type
        self.bus = bus
        self.config = config or {}
        self.reset()

    def reset(self):
        try:
            from ovos_utils.ocp import PlaybackMode
        except ImportError as e:
            raise RuntimeError("This method requires ovos-utils ~=0.1") from e
        self.active_skills = {}
        self.active_skills_lock = Lock()
        self.query_replies = []
        self.searching = False
        self.search_start = 0
        self.query_timeouts = self.config.get("min_timeout", 5)
        if self.config.get("playback_mode") in [PlaybackMode.AUDIO_ONLY]:
            self.has_gui = False
        else:
            self.has_gui = is_gui_running() or is_gui_connected(self.bus)

    @_ensure_message_kwarg()
    def send(self, skill_id: str = None, source_message: Optional[Message] = None):
        self.query_replies = []
        self.query_timeouts = self.config.get("min_timeout", 5)
        self.search_start = time.time()
        self.searching = True
        self.register_events()
        if skill_id:
            self.bus.emit(source_message.forward(f'ovos.common_play.query.{skill_id}',
                                                 {"phrase": self.query,
                                                  "question_type": self.media_type}))
        else:
            self.bus.emit(source_message.forward('ovos.common_play.query',
                                                 {"phrase": self.query,
                                                  "question_type": self.media_type}))

    def wait(self):
        try:
            from ovos_utils.ocp import MediaType
        except ImportError as e:
            raise RuntimeError("This method requires ovos-utils ~=0.1") from e
        # if there is no match type defined, lets increase timeout a bit
        # since all skills need to search
        if self.media_type == MediaType.GENERIC:
            timeout = self.config.get("max_timeout", 15) + 3  # timeout bonus
        else:
            timeout = self.config.get("max_timeout", 15)
        while self.searching and time.time() - self.search_start <= timeout:
            time.sleep(0.1)
        self.searching = False
        self.remove_events()

    @property
    def results(self) -> List[dict]:
        return [s for s in self.query_replies if s.get("results")]

    def register_events(self):
        LOG.debug("Registering Search Bus Events")
        self.bus.on("ovos.common_play.skill.search_start", self.handle_skill_search_start)
        self.bus.on("ovos.common_play.skill.search_end", self.handle_skill_search_end)
        self.bus.on("ovos.common_play.query.response", self.handle_skill_response)

    def remove_events(self):
        LOG.debug("Removing Search Bus Events")
        self.bus.remove_all_listeners("ovos.common_play.skill.search_start")
        self.bus.remove_all_listeners("ovos.common_play.skill.search_end")
        self.bus.remove_all_listeners("ovos.common_play.query.response")

    def handle_skill_search_start(self, message):
        skill_id = message.data["skill_id"]
        LOG.debug(f"{message.data['skill_id']} is searching")
        with self.active_skills_lock:
            if skill_id not in self.active_skills:
                self.active_skills[skill_id] = Lock()

    def handle_skill_response(self, message):
        search_phrase = message.data["phrase"]
        if search_phrase != self.query:
            # not an answer for this search query
            return
        timeout = message.data.get("timeout")
        skill_id = message.data['skill_id']
        # LOG.debug(f"OVOSCommonPlay result: {skill_id}")

        # in case this handler fires before the search start handler
        with self.active_skills_lock:
            if skill_id not in self.active_skills:
                self.active_skills[skill_id] = Lock()
        with self.active_skills[skill_id]:
            if message.data.get("searching"):
                # extend the timeout by N seconds
                if timeout and self.config.get("allow_extensions", True):
                    self.query_timeouts += timeout
                # else -> expired search

            else:
                # Collect replies until the timeout
                if not self.searching and not len(self.query_replies):
                    LOG.debug("  too late!! ignored in track selection process")
                    LOG.warning(f"{skill_id} is not answering fast enough!")
                    return

                # populate search playlist
                res = message.data.get("results", [])
                LOG.debug(f'got {len(res)} results from {skill_id}')
                if res:
                    self.query_replies.append(message.data)

                    # abort searching if we gathered enough results
                    # TODO ensure we have a decent confidence match, if all matches
                    #  are < 50% conf extend timeout instead
                    if time.time() - self.search_start > self.query_timeouts:
                        if self.searching:
                            self.searching = False
                            LOG.debug("common play query timeout, parsing results")

                    elif self.searching:
                        for res in message.data.get("results", []):
                            if res.get("match_confidence", 0) >= \
                                    self.config.get("early_stop_thresh", 85):
                                # got a really good match, dont search further
                                LOG.info(
                                    "Receiving very high confidence match, stopping "
                                    "search early")

                                # allow other skills to "just miss"
                                early_stop_grace = \
                                    self.config.get("early_stop_grace_period", 0.5)
                                if early_stop_grace:
                                    LOG.debug(
                                        f"  - grace period: {early_stop_grace} seconds")
                                    time.sleep(early_stop_grace)
                                self.searching = False
                                return

    def handle_skill_search_end(self, message):
        skill_id = message.data["skill_id"]
        LOG.debug(f"{message.data['skill_id']} finished search")
        with self.active_skills_lock:
            if skill_id in self.active_skills:
                with self.active_skills[skill_id]:
                    del self.active_skills[skill_id]

        # if this was the last skill end searching period
        time.sleep(0.5)
        # TODO this sleep is hacky, but avoids a race condition in
        # case some skill just decides to respond before the others even
        # acknowledge search is starting, this gives more than enough time
        # for self.active_skills to be populated, a better approach should
        # be employed but this works fine for now
        if not self.active_skills and self.searching:
            LOG.info("Received search responses from all skills!")
            self.searching = False


##########################################################
# WIP ZONE - APIs below used for ovos-media


class OCPAudioServiceInterface:
    """Internal OCP audio subsystem
    most likely you should use OCPInterface instead
    NOTE: this class operates with uris not with MediaEntry/Playlist/dict entries
    """

    def __init__(self, bus=None):
        self.bus = bus or get_mycroft_bus()

    def play(self, tracks=None, utterance=None, repeat=None):
        """Start playback.

        Args:
            tracks: track uri or list of track uri's
                    Each track can be added as a tuple with (uri, mime)
                    to give a hint of the mime type to the system
            utterance: forward utterance for further processing by the
                        audio service.
            repeat: if the playback should be looped
        """
        repeat = repeat or False
        tracks = tracks or []
        utterance = utterance or ''
        if isinstance(tracks, (str, tuple)):
            tracks = [tracks]
        elif not isinstance(tracks, list):
            raise ValueError
        tracks = [ensure_uri(t) for t in tracks]
        self.bus.emit(Message('ovos.audio.service.play',
                              data={'tracks': tracks,
                                    'utterance': utterance,
                                    'repeat': repeat}))

    def stop(self):
        """Stop the track."""
        self.bus.emit(Message('ovos.audio.service.stop'))

    def next(self):
        """Change to next track."""
        self.bus.emit(Message('ovos.audio.service.next'))

    def prev(self):
        """Change to previous track."""
        self.bus.emit(Message('ovos.audio.service.prev'))

    def pause(self):
        """Pause playback."""
        self.bus.emit(Message('ovos.audio.service.pause'))

    def resume(self):
        """Resume paused playback."""
        self.bus.emit(Message('ovos.audio.service.resume'))

    def get_track_length(self):
        """
        getting the duration of the audio in seconds
        """
        length = 0
        info = self.bus.wait_for_response(
            Message('ovos.audio.service.get_track_length'),
            timeout=1)
        if info:
            length = info.data.get("length") or 0
        return length / 1000  # convert to seconds

    def get_track_position(self):
        """
        get current position in seconds
        """
        pos = 0
        info = self.bus.wait_for_response(
            Message('ovos.audio.service.get_track_position'),
            timeout=1)
        if info:
            pos = info.data.get("position") or 0
        return pos / 1000  # convert to seconds

    def set_track_position(self, seconds):
        """Seek X seconds.

        Arguments:
            seconds (int): number of seconds to seek, if negative rewind
        """
        self.bus.emit(Message('ovos.audio.service.set_track_position',
                              {"position": seconds * 1000}))  # convert to ms

    def seek(self, seconds: Union[int, float, timedelta] = 1):
        """Seek X seconds.

        Args:
            seconds (int): number of seconds to seek, if negative rewind
        """
        if isinstance(seconds, timedelta):
            seconds = seconds.total_seconds()
        if seconds < 0:
            self.seek_backward(abs(seconds))
        else:
            self.seek_forward(seconds)

    def seek_forward(self, seconds: Union[int, float, timedelta] = 1):
        """Skip ahead X seconds.

        Args:
            seconds (int): number of seconds to skip
        """
        if isinstance(seconds, timedelta):
            seconds = seconds.total_seconds()
        self.bus.emit(Message('ovos.audio.service.seek_forward',
                              {"seconds": seconds}))

    def seek_backward(self, seconds: Union[int, float, timedelta] = 1):
        """Rewind X seconds

         Args:
            seconds (int): number of seconds to rewind
        """
        if isinstance(seconds, timedelta):
            seconds = seconds.total_seconds()
        self.bus.emit(Message('ovos.audio.service.seek_backward',
                              {"seconds": seconds}))

    def track_info(self):
        """Request information of current playing track.

        Returns:
            Dict with track info.
        """
        info = self.bus.wait_for_response(
            Message('ovos.audio.service.track_info'),
            reply_type='ovos.audio.service.track_info_reply',
            timeout=1)
        return info.data if info else {}

    def available_backends(self):
        """Return available audio backends.

        Returns:
            dict with backend names as keys
        """
        msg = Message('ovos.audio.service.list_backends')
        response = self.bus.wait_for_response(msg)
        return response.data if response else {}

    @property
    def is_playing(self):
        """True if the audioservice is playing, else False."""
        return self.track_info() != {}


class OCPVideoServiceInterface:
    """Internal OCP video subsystem
    most likely you should use OCPInterface instead
    NOTE: this class operates with uris not with MediaEntry/Playlist/dict entries
    """

    def __init__(self, bus=None):
        self.bus = bus or get_mycroft_bus()

    def play(self, tracks=None, utterance=None, repeat=None):
        """Start playback.

        Args:
            tracks: track uri or list of track uri's
                    Each track can be added as a tuple with (uri, mime)
                    to give a hint of the mime type to the system
            utterance: forward utterance for further processing by the
                        video service.
            repeat: if the playback should be looped
        """
        repeat = repeat or False
        tracks = tracks or []
        utterance = utterance or ''
        if isinstance(tracks, (str, tuple)):
            tracks = [tracks]
        elif not isinstance(tracks, list):
            raise ValueError
        tracks = [ensure_uri(t) for t in tracks]
        self.bus.emit(Message('ovos.video.service.play',
                              data={'tracks': tracks,
                                    'utterance': utterance,
                                    'repeat': repeat}))

    def stop(self):
        """Stop the track."""
        self.bus.emit(Message('ovos.video.service.stop'))

    def next(self):
        """Change to next track."""
        self.bus.emit(Message('ovos.video.service.next'))

    def prev(self):
        """Change to previous track."""
        self.bus.emit(Message('ovos.video.service.prev'))

    def pause(self):
        """Pause playback."""
        self.bus.emit(Message('ovos.video.service.pause'))

    def resume(self):
        """Resume paused playback."""
        self.bus.emit(Message('ovos.video.service.resume'))

    def get_track_length(self):
        """
        getting the duration of the video in seconds
        """
        length = 0
        info = self.bus.wait_for_response(
            Message('ovos.video.service.get_track_length'),
            timeout=1)
        if info:
            length = info.data.get("length") or 0
        return length / 1000  # convert to seconds

    def get_track_position(self):
        """
        get current position in seconds
        """
        pos = 0
        info = self.bus.wait_for_response(
            Message('ovos.video.service.get_track_position'),
            timeout=1)
        if info:
            pos = info.data.get("position") or 0
        return pos / 1000  # convert to seconds

    def set_track_position(self, seconds):
        """Seek X seconds.

        Arguments:
            seconds (int): number of seconds to seek, if negative rewind
        """
        self.bus.emit(Message('ovos.video.service.set_track_position',
                              {"position": seconds * 1000}))  # convert to ms

    def seek(self, seconds=1):
        """Seek X seconds.

        Args:
            seconds (int): number of seconds to seek, if negative rewind
        """
        if isinstance(seconds, timedelta):
            seconds = seconds.total_seconds()
        if seconds < 0:
            self.seek_backward(abs(seconds))
        else:
            self.seek_forward(seconds)

    def seek_forward(self, seconds: Union[int, float, timedelta] = 1):
        """Skip ahead X seconds.

        Args:
            seconds (int): number of seconds to skip
        """
        if isinstance(seconds, timedelta):
            seconds = seconds.total_seconds()
        self.bus.emit(Message('ovos.video.service.seek_forward',
                              {"seconds": seconds}))

    def seek_backward(self, seconds: Union[int, float, timedelta] = 1):
        """Rewind X seconds

         Args:
            seconds (int): number of seconds to rewind
        """
        if isinstance(seconds, timedelta):
            seconds = seconds.total_seconds()
        self.bus.emit(Message('ovos.video.service.seek_backward',
                              {"seconds": seconds}))

    def track_info(self):
        """Request information of current playing track.

        Returns:
            Dict with track info.
        """
        info = self.bus.wait_for_response(
            Message('ovos.video.service.track_info'),
            reply_type='ovos.video.service.track_info_reply',
            timeout=1)
        return info.data if info else {}

    def available_backends(self):
        """Return available video backends.

        Returns:
            dict with backend names as keys
        """
        msg = Message('ovos.video.service.list_backends')
        response = self.bus.wait_for_response(msg)
        return response.data if response else {}

    @property
    def is_playing(self):
        """True if the videoservice is playing, else False."""
        return self.track_info() != {}


class OCPWebServiceInterface:
    """Internal OCP web view subsystem
    most likely you should use OCPInterface instead
    NOTE: this class operates with uris not with MediaEntry/Playlist/dict entries
    """

    def __init__(self, bus=None):
        self.bus = bus or get_mycroft_bus()

    def play(self, tracks=None, utterance=None, repeat=None):
        """Start playback.

        Args:
            tracks: track uri or list of track uri's
                    Each track can be added as a tuple with (uri, mime)
                    to give a hint of the mime type to the system
            utterance: forward utterance for further processing by the
                        web service.
            repeat: if the playback should be looped
        """
        repeat = repeat or False
        tracks = tracks or []
        utterance = utterance or ''
        if isinstance(tracks, (str, tuple)):
            tracks = [tracks]
        elif not isinstance(tracks, list):
            raise ValueError
        tracks = [ensure_uri(t) for t in tracks]
        self.bus.emit(Message('ovos.web.service.play',
                              data={'tracks': tracks,
                                    'utterance': utterance,
                                    'repeat': repeat}))

    def stop(self):
        """Stop the track."""
        self.bus.emit(Message('ovos.web.service.stop'))

    def next(self):
        """Change to next track."""
        self.bus.emit(Message('ovos.web.service.next'))

    def prev(self):
        """Change to previous track."""
        self.bus.emit(Message('ovos.web.service.prev'))

    def pause(self):
        """Pause playback."""
        self.bus.emit(Message('ovos.web.service.pause'))

    def resume(self):
        """Resume paused playback."""
        self.bus.emit(Message('ovos.web.service.resume'))

    def get_track_length(self):
        """
        getting the duration of the web in seconds
        """
        length = 0
        info = self.bus.wait_for_response(
            Message('ovos.web.service.get_track_length'),
            timeout=1)
        if info:
            length = info.data.get("length") or 0
        return length / 1000  # convert to seconds

    def get_track_position(self):
        """
        get current position in seconds
        """
        pos = 0
        info = self.bus.wait_for_response(
            Message('ovos.web.service.get_track_position'),
            timeout=1)
        if info:
            pos = info.data.get("position") or 0
        return pos / 1000  # convert to seconds

    def set_track_position(self, seconds):
        """Seek X seconds.

        Arguments:
            seconds (int): number of seconds to seek, if negative rewind
        """
        self.bus.emit(Message('ovos.web.service.set_track_position',
                              {"position": seconds * 1000}))  # convert to ms

    def seek(self, seconds: Union[int, float, timedelta] = 1):
        """Seek X seconds.

        Args:
            seconds (int): number of seconds to seek, if negative rewind
        """
        if isinstance(seconds, timedelta):
            seconds = seconds.total_seconds()
        if seconds < 0:
            self.seek_backward(abs(seconds))
        else:
            self.seek_forward(seconds)

    def seek_forward(self, seconds: Union[int, float, timedelta] = 1):
        """Skip ahead X seconds.

        Args:
            seconds (int): number of seconds to skip
        """
        if isinstance(seconds, timedelta):
            seconds = seconds.total_seconds()
        self.bus.emit(Message('ovos.web.service.seek_forward',
                              {"seconds": seconds}))

    def seek_backward(self, seconds: Union[int, float, timedelta] = 1):
        """Rewind X seconds

         Args:
            seconds (int): number of seconds to rewind
        """
        if isinstance(seconds, timedelta):
            seconds = seconds.total_seconds()
        self.bus.emit(Message('ovos.web.service.seek_backward',
                              {"seconds": seconds}))

    def track_info(self):
        """Request information of current playing track.

        Returns:
            Dict with track info.
        """
        info = self.bus.wait_for_response(
            Message('ovos.web.service.track_info'),
            reply_type='ovos.web.service.track_info_reply',
            timeout=1)
        return info.data if info else {}

    def available_backends(self):
        """Return available web backends.

        Returns:
            dict with backend names as keys
        """
        msg = Message('ovos.web.service.list_backends')
        response = self.bus.wait_for_response(msg)
        return response.data if response else {}

    @property
    def is_playing(self):
        """True if the webservice is playing, else False."""
        return self.track_info() != {}



# Copyright 2019 Mycroft AI Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""
Tools and constructs that are useful together with the messagebus.
"""
import orjson

from ovos_config.config import read_mycroft_config
from ovos_config.locale import get_default_lang
from ovos_utils.json_helper import merge_dict
from ovos_bus_client import MessageBusClient
from ovos_bus_client.message import dig_for_message, Message
from ovos_bus_client.session import SessionManager
from ovos_bus_client.util.scheduler import EventScheduler
from ovos_bus_client.util.utils import create_echo_function

_DEFAULT_WS_CONFIG = {"host": "0.0.0.0",
                      "port": 8181,
                      "route": "/core",
                      "ssl": False}


def get_message_lang(message=None):
    message = message or dig_for_message()
    if not message:
        return None
    # old style lang param
    lang = message.data.get("lang") or message.context.get("lang")
    if lang:
        return lang

    # new style session lang
    if "session_id" in message.context or "session" in message.context:
        sess = SessionManager.get(message)
        return sess.lang

    return get_default_lang()


def get_websocket(host, port, route='/', ssl=False, threaded=True):
    """
    Returns a connection to a websocket
    """

    client = MessageBusClient(host, port, route, ssl)
    if threaded:
        client.run_in_thread()
    return client


def get_mycroft_bus(host: str = None, port: int = None, route: str = None,
                    ssl: bool = None):
    """
    Returns a connection to the mycroft messagebus
    """
    config = read_mycroft_config().get('websocket') or dict()
    host = host or config.get('host') or _DEFAULT_WS_CONFIG['host']
    port = port or config.get('port') or _DEFAULT_WS_CONFIG['port']
    route = route or config.get('route') or _DEFAULT_WS_CONFIG['route']
    if ssl is None:
        ssl = config.get('ssl') if 'ssl' in config else \
            _DEFAULT_WS_CONFIG['ssl']
    return get_websocket(host, port, route, ssl)


def listen_for_message(msg_type, handler, bus=None):
    """
    Continuously listens and reacts to a specific messagetype on the mycroft messagebus

    NOTE: when finished you should call bus.remove(msg_type, handler)
    """
    bus = bus or get_mycroft_bus()
    bus.on(msg_type, handler)
    return bus


def listen_once_for_message(msg_type, handler, bus=None):
    """
    listens and reacts once to a specific messagetype on the mycroft messagebus
    """
    auto_close = bus is None
    bus = bus or get_mycroft_bus()

    def _handler(message):
        handler(message)
        if auto_close:
            bus.close()

    bus.once(msg_type, _handler)
    return bus


def wait_for_reply(message, reply_type=None, timeout=3.0, bus=None):
    """Send a message and wait for a response.

    Args:
        message (FakeMessage or str or dict): message object or type to send
        reply_type (str): the message type of the expected reply.
                          Defaults to "<message.type>.response".
        timeout: seconds to wait before timeout, defaults to 3
    Returns:
        The received message or None if the response timed out
    """
    auto_close = bus is None
    bus = bus or get_mycroft_bus()
    if isinstance(message, str):
        try:
            message = orjson.loads(message)
        except:
            pass
    if isinstance(message, str):
        message = Message(message)
    elif isinstance(message, dict):
        message = Message(message["type"],
                          message.get("data"),
                          message.get("context"))
    elif not isinstance(message, Message):
        raise ValueError
    response = bus.wait_for_response(message, reply_type, timeout)
    if auto_close:
        bus.close()
    return response


def send_message(message, data=None, context=None, bus=None):
    auto_close = bus is None
    bus = bus or get_mycroft_bus()
    if isinstance(message, str):
        if isinstance(data, dict) or isinstance(context, dict):
            message = Message(message, data, context)
        else:
            try:
                message = orjson.loads(message)
            except:
                message = Message(message)
    if isinstance(message, dict):
        message = Message(message["type"],
                          message.get("data"),
                          message.get("context"))
    if not isinstance(message, Message):
        raise ValueError
    bus.emit(message)
    if auto_close:
        bus.close()


def send_binary_data_message(binary_data, msg_type="mycroft.binary.data",
                             msg_data=None, msg_context=None, bus=None):
    msg_data = msg_data or {}
    msg = {
        "type": msg_type,
        "data": merge_dict(msg_data, {"binary": binary_data.hex()}),
        "context": msg_context or None
    }
    send_message(msg, bus=bus)


def send_binary_file_message(filepath, msg_type="mycroft.binary.file",
                             msg_context=None, bus=None):
    with open(filepath, 'rb') as f:
        binary_data = f.read()
    msg_data = {"path": filepath}
    send_binary_data_message(binary_data, msg_type=msg_type, msg_data=msg_data,
                             msg_context=msg_context, bus=bus)


def decode_binary_message(message):
    if isinstance(message, str):
        try:  # json string
            message = orjson.loads(message)
            binary_data = message.get("binary") or message["data"]["binary"]
        except:  # hex string
            binary_data = message
    elif isinstance(message, dict):
        # data field or serialized message
        binary_data = message.get("binary") or message["data"]["binary"]
    else:
        # message object
        binary_data = message.data["binary"]
    # decode hex string
    return bytearray.fromhex(binary_data)



# Copyright 2017 Mycroft AI Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""
Classes and functions related to the Mycroft Message.

The Message object is the core construct passed on the message bus
it contains methods for tracking message context and
serializing / deserializing the message for transmission.
"""

import inspect
import orjson
import re
from copy import deepcopy
from typing import Optional
from binascii import hexlify, unhexlify
from ovos_utils.log import LOG, deprecated
from ovos_utils.security import encrypt, decrypt
from ovos_config.config import Configuration

try:
    from lingua_franca.parse import normalize
except ImportError:
    # optional LF import
    def normalize(text, *args, **kwargs):
        return text

try:
    from mycroft_bus_client.message import Message as _MsgBase, \
        CollectionMessage as _CollectionMsgBase

except ImportError:

    # TODO - code in the wild does isinstance checks
    # this conditional subclassing should be removed ASAP, it is only here for the migration period
    # mycroft_bus_client is abandonware until further notice from MycroftAI

    class _MsgBase:
        pass


    class _CollectionMsgBase(_MsgBase):
        pass


class _MessageMeta(type):
    """ To override isinstance checks we need to use a metaclass """

    def __instancecheck__(self, instance):
        try:
            from mycroft_bus_client.message import Message as _MycroftMessage
            return isinstance(instance, _MycroftMessage) or \
                super().__instancecheck__(instance)
        except:
            return super().__instancecheck__(instance)


class Message(_MsgBase, metaclass=_MessageMeta):
    """Holds and manipulates data sent over the websocket

        Message objects will be used to send information back and forth
        between processes of Mycroft.

    Attributes:
        msg_type (str): type of data sent within the message.
        data (dict): data sent within the message
        context: info about the message not part of data such as source,
            destination or domain.
    """
    # if set all messages are AES encrypted
    _secret_key = Configuration().get("websocket", {}).get("secret_key")
    # if set to False, will refuse to deserialize unencrypted messages for processing
    _allow_unencrypted = Configuration().get("websocket", {}).get("allow_unencrypted", _secret_key is None)

    def __init__(self, msg_type: str, data: dict = None, context: dict = None):
        """Used to construct a message object

        Message objects will be used to send information back and forth
        between processes of mycroft service, voice, skill and cli
        """
        data = data or {}
        context = context or {}
        assert isinstance(data, dict)
        assert isinstance(context, dict)
        self.msg_type = msg_type
        self.data = data
        self.context = context

    def __eq__(self, other):
        if not isinstance(other, Message):
            return False
        return other.msg_type == self.msg_type and \
            other.data == self.data and \
            other.context == self.context

    def serialize(self) -> str:
        """This returns a string of the message info.

        This makes it easy to send over a websocket. This uses
        json dumps to generate the string with type, data and context

        Returns:
            str: a json string representation of the message.
        """
        # handle Session and Message objects
        data = self._json_dump(self.data)
        ctxt = self._json_dump(self.context)

        msg = orjson.dumps({'type': self.msg_type, 'data': data, 'context': ctxt}).decode("utf-8")
        if self._secret_key:
            payload = encrypt_as_dict(self._secret_key, msg)
            return orjson.dumps(payload).decode("utf-8")
        return msg

    @property
    def as_dict(self) -> dict:
        return orjson.loads(self.serialize())

    @staticmethod
    def _json_dump(value):

        from ovos_bus_client.apis.gui import _GUIDict

        def serialize_item(x):
            try:
                if hasattr(x, "serialize"):
                    return x.serialize()
            except:
                pass
            if isinstance(x, list):
                for idx, it in enumerate(x):
                    x[idx] = serialize_item(it)
            if isinstance(x, dict) and not isinstance(x, _GUIDict):
                for k, v in x.items():
                    x[k] = serialize_item(v)
            return x
        assert isinstance(value, dict)
        data = {k: serialize_item(v) for k, v in value.items()}
        return data

    @staticmethod
    def _json_load(value):
        if isinstance(value, str):
            obj = orjson.loads(value)
        else:
            obj = value
        assert isinstance(obj, dict)
        if Message._secret_key:
            if 'ciphertext' in obj:
                obj = decrypt_from_dict(Message._secret_key, obj)
            elif not Message._allow_unencrypted:
                raise RuntimeError("got an unencrypted message, configured to refuse")
        return obj

    @staticmethod
    def deserialize(value: str) -> _MsgBase:
        """
        This takes a string and constructs a message object.

        This makes it easy to take strings from the websocket and create
        a message object.  This uses json loads to get the info and generate
        the message object.

        Args:
            value(str): This is the json string received from the websocket

        Returns:
            Message: message object constructed from the json string passed
            int the function.
            value(str): This is the string received from the websocket
        """
        obj = Message._json_load(value)
        return Message(obj.get('type') or '',
                       obj.get('data') or {},
                       obj.get('context') or {})

    def forward(self, msg_type: str, data: dict = None) -> _MsgBase:
        """
        Keep context and forward message

        This will take the same parameters as a message object but use
        the current message object as a reference.  It will copy the context
        from the existing message object.

        Args:
            msg_type (str): type of message
            data (dict): data for message

        Returns:
            Message: Message object to be used on the reply to the message
        """
        data = data or {}
        return Message(msg_type, data, context=self.context)

    def reply(self, msg_type: str, data: dict = None,
              context: dict = None) -> _MsgBase:
        """
        Construct a reply message for a given message

        This will take the same parameters as a message object but use
        the current message object as a reference.  It will copy the context
        from the existing message object and add any context passed in to
        the function.  Check for a destination passed in to the function from
        the data object and add that to the context as a destination.  If the
        context has a source then that will be swapped with the destination
        in the context.  The new message will then have data passed in plus the
        new context generated.

        Args:
            msg_type (str): type of message
            data (dict): data for message
            context: intended context for new message

        Returns:
            Message: Message object to be used on the reply to the message
        """
        data = deepcopy(data) or {}
        context = context or {}

        new_context = deepcopy(self.context)
        for key in context:
            new_context[key] = context[key]
        if 'destination' in data:
            new_context['destination'] = data['destination']
        if 'source' in new_context and 'destination' in new_context:
            s = new_context['destination']
            new_context['destination'] = new_context['source']
            new_context['source'] = s
        return Message(msg_type, data, context=new_context)

    def response(self, data: dict = None, context: dict = None) -> _MsgBase:
        """
        Construct a response message for the message

        Constructs a reply with the data and appends the expected
        ".response" to the message

        Args:
            data (dict): message data
            context (dict): message context
        Returns
            (Message) message with the type modified to match default response
        """
        return self.reply(self.msg_type + '.response', data, context)

    def publish(self, msg_type: str, data: dict,
                context: dict = None) -> _MsgBase:
        """
        Copy the original context and add passed in context.  Delete
        any target in the new context. Return a new message object with
        passed in data and new context.  Type remains unchanged.

        Args:
            msg_type (str): type of message
            data (dict): data to send with message
            context: context added to existing context

        Returns:
            Message: Message object to publish
        """
        context = context or {}
        new_context = self.context.copy()
        for key in context:
            new_context[key] = context[key]

        if 'target' in new_context:
            del new_context['target']

        return Message(msg_type, data, context=new_context)

    @deprecated("This method is deprecated with no replacement", "0.1.0")
    def utterance_remainder(self):
        """
        DEPRECATED - mycroft-core hack, used by some skills in the wild

        For intents get the portion not consumed by Adapt.

        For example: if they say 'Turn on the family room light' and there are
        entity matches for "turn on" and "light", then it will leave behind
        " the family room " which is then normalized to "family room".

        Returns:
            str: Leftover words or None if not an utterance.
        """
        utt = normalize(self.data.get("utterance", ""))
        if utt and "__tags__" in self.data:
            for token in self.data["__tags__"]:
                # Substitute only whole words matching the token
                utt = re.sub(r'\b' + token.get("key", "") + r"\b", "", utt)
        return normalize(utt)


def encrypt_as_dict(key: str, data: str, nonce=None) -> dict:
    ciphertext, tag, nonce = encrypt(key, data, nonce=nonce)
    return {"ciphertext": hexlify(ciphertext).decode('utf-8'),
            "tag": hexlify(tag).decode('utf-8'),
            "nonce": hexlify(nonce).decode('utf-8')}


def decrypt_from_dict(key: str, data: dict) -> str:
    ciphertext = unhexlify(data["ciphertext"])
    if data.get("tag") is None:  # web crypto
        ciphertext, tag = ciphertext[:-16], ciphertext[-16:]
    else:
        tag = unhexlify(data["tag"])
    nonce = unhexlify(data["nonce"])
    return decrypt(key, ciphertext, tag, nonce)


def dig_for_message(max_records: int = 10) -> Optional[Message]:
    """
    Dig Through the stack for message. Looks at the current stack
    for a passed argument of type 'Message'.
    Args:
        max_records (int): Maximum number of stack records to look through

    Returns:
        Message if found in args, else None
    """
    stack = inspect.stack()[1:]  # First frame will be this function call
    stack = stack if len(stack) <= max_records else stack[:max_records]
    for record in stack:
        args = inspect.getargvalues(record.frame)
        if args.args:
            for arg in args.args:
                if isinstance(args.locals[arg], Message):
                    return args.locals[arg]
    return None


class CollectionMessage(Message, _CollectionMsgBase):
    """Extension of the Message class for use with collect handlers.

    The class provides the convenience methods success and failure to report
    these states back to the origin.
    """

    def __init__(self, msg_type, handler_id, query_id, data=None, context=None):
        super().__init__(msg_type, data, context)
        self.handler_id = handler_id
        self.query_id = query_id

    @classmethod
    def from_message(cls, message, handler_id, query_id):
        """Build a CollectionMessage based of a Message object.

        Args:
            message (Message): the original message
            handler_id (str): the handler_id of the recipient
            query_id (str): the query session id

        Returns:
            CollectionMessage based on the original Message object
        """
        return cls(message.msg_type, handler_id, query_id,
                   message.data, message.context)

    def success(self, data=None, context=None):
        """Create a message indicating a successful result.

        The handler could handle the query and created some sort of response.
        The source and destination is switched in the context like when
        sending a normal response message.

            data (dict): message data
            context (dict): message context
        Returns:
            Message
        """
        data = data or {}
        data['query'] = self.query_id
        data['handler'] = self.handler_id
        data['succeeded'] = True
        response_message = self.reply(self.msg_type + '.response',
                                      data,
                                      context or self.context)
        return response_message

    def failure(self):
        """Create a message indicating a failing result.

        The handler could not handle the query.
        The source and destination is switched in the context like when
        sending a normal response message.

            data (dict): message data
            context (dict): message context
        Returns:
            Message
        """
        data = {}
        data['query'] = self.query_id
        data['handler'] = self.handler_id
        data['succeeded'] = False
        response_message = self.reply(self.msg_type + '.response',
                                      data,
                                      self.context)
        return response_message

    def extend(self, timeout):
        """Extend current timeout,

        The timeout provided will be added to the existing timeout.
        The source and destination is switched in the context like when
        sending a normal response message.

        Arguments:
            timeout (int/float): timeout extension

        Returns:
            Extension message.
        """
        data = {}
        data['query'] = self.query_id
        data['handler'] = self.handler_id
        data['timeout'] = timeout
        response_message = self.reply(self.msg_type + '.handling',
                                      data,
                                      self.context)
        return response_message


class GUIMessage(Message):
    def __init__(self, msg_type, **kwargs):
        super().__init__(msg_type, data=kwargs)

    def serialize(self) -> str:
        """This returns a string of the message info.

        This makes it easy to send over a websocket. This uses
        json dumps to generate the string with type, data and context

        Returns:
            str: a json string representation of the message.
        """
        data = self._json_dump(self.data)
        msg = orjson.dumps({'type': self.msg_type, **data}).decode("utf-8")
        if self._secret_key:
            payload = encrypt_as_dict(self._secret_key, msg)
            return orjson.dumps(payload).decode("utf-8")
        return msg

    @staticmethod
    def deserialize(value: str):
        value = Message._json_load(value)
        msg_type = value.pop("type")
        return GUIMessage(msg_type, **value)



# Copyright 2019 Mycroft AI Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""
The scheduler module allows setting up scheduled messages.

A scheduled message will be kept and not sent until the system clock time
criteria is met.
"""

import json
import shutil
import time

from typing import Optional
from threading import Event
from os.path import isfile, join, expanduser
from threading import Thread, Lock

from ovos_config.config import Configuration
from ovos_config.locations import get_xdg_data_save_path, get_xdg_config_save_path
from ovos_utils.log import LOG, log_deprecation
from ovos_utils.events import EventContainer as _EventContainer
from ovos_utils.events import EventSchedulerInterface as _SchedulerInterface
from ovos_bus_client.message import Message


def repeat_time(sched_time: float, repeat: float) -> float:
    """
    Next scheduled time for repeating event. Guarantees that the
    time is not in the past (but could skip interim events)

    Args:
        sched_time (float): Scheduled unix time for the event
        repeat (float):     Repeat period in seconds

    Returns: (float) time for next event
    """
    next_time = sched_time + repeat
    while next_time < time.time():
        # Schedule at an offset to assure no doubles
        next_time = time.time() + abs(repeat)
    return next_time


class EventScheduler(Thread):
    """
    Create an event scheduler thread. Will send messages at a
    predetermined time to the registered targets.

    Arguments:
        bus:            Mycroft messagebus client
        schedule_file:  filename used to store pending events to on
                        shutdown. File is created in XDG_DATA_HOME
        autostart: if True, start scheduler on init
    """

    def __init__(self, bus,
                 schedule_file: str = 'schedule.json', autostart: bool = True):
        super().__init__()

        self.events = {}
        self.event_lock = Lock()

        self.bus = bus

        core_conf = Configuration()
        data_dir = core_conf.get('data_dir') or get_xdg_data_save_path()
        old_schedule_path = join(expanduser(data_dir), schedule_file)

        self.schedule_file = join(get_xdg_config_save_path(), schedule_file)
        if isfile(old_schedule_path):
            shutil.move(old_schedule_path, self.schedule_file)

        if self.schedule_file:
            self.load()

        self.bus.on('mycroft.scheduler.schedule_event',
                    self.schedule_event_handler)
        self.bus.on('mycroft.scheduler.remove_event',
                    self.remove_event_handler)
        self.bus.on('mycroft.scheduler.update_event',
                    self.update_event_handler)
        self.bus.on('mycroft.scheduler.get_event',
                    self.get_event_handler)

        self._running = Event()
        self._stopping = Event()
        if autostart:
            self.start()
            self._running.wait(10)
        else:
            # Explicitly define event states
            self._stopping.set()
            self._running.clear()

    @property
    def is_running(self) -> bool:
        """
        Return True while scheduler is running
        """
        return not self._stopping.is_set()

    @is_running.setter
    def is_running(self, value: bool):
        if value is True:
            self._stopping.clear()
        elif value is False:
            self._stopping.set()

    def load(self):
        """
        Load json data with active events from json file.
        """
        if isfile(self.schedule_file):
            json_data = {}
            with open(self.schedule_file) as schedule_file:
                try:
                    json_data = json.load(schedule_file)
                except Exception as exc:
                    LOG.error(exc)
            current_time = time.time()
            with self.event_lock:
                for key in json_data:
                    event_list = json_data[key]
                    # discard non repeating events that has already happened
                    self.events[key] = [tuple(evt) for evt in event_list
                                        if evt[0] > current_time or evt[1]]

    def run(self):
        """
        Check events periodically until stopped
        """
        LOG.info("EventScheduler Started")
        self._stopping.clear()
        self._running.set()
        while not self._stopping.wait(0.5):
            try:
                self.check_state()
            except Exception as e:
                LOG.exception(e)
        LOG.info("EventScheduler Stopped")

    def check_state(self):
        """
        Check if any event should be triggered.
        """
        with self.event_lock:
            # Check all events
            pending_messages = []
            for event in self.events:
                current_time = time.time()
                e = self.events[event]
                # Get scheduled times that has passed
                passed = ((t, r, d, c) for
                          (t, r, d, c) in e if t <= current_time)
                # and remaining times that we're still waiting for
                remaining = [(t, r, d, c) for
                             t, r, d, c in e if t > current_time]
                # Trigger registered methods
                for sched_time, repeat, data, context in passed:
                    pending_messages.append(Message(event, data, context))
                    # if this is a repeated event add a new trigger time
                    if repeat:
                        next_time = repeat_time(sched_time, repeat)
                        remaining.append((next_time, repeat, data, context))
                # update list of events
                self.events[event] = remaining

        # Remove events that are now completed
        self.clear_empty()

        # Finally, emit the queued up events that triggered
        for msg in pending_messages:
            LOG.debug(f"Call scheduled event: {msg.msg_type}")
            self.bus.emit(msg)

    def schedule_event(self, event: str, sched_time: float,
                       repeat: Optional[float] = None,
                       data: Optional[dict] = None,
                       context: Optional[dict] = None):
        """
        Add event to pending event schedule.

        Arguments:
            event (str): Handler for the event
            sched_time (float): epoch time of event
            repeat ([type], optional): Defaults to None. [description]
            data ([type], optional): Defaults to None. [description]
            context (dict, optional): context (dict, optional): message
                                      context to send when the
                                      handler is called
        """
        data = data or {}
        with self.event_lock:
            # get current list of scheduled times for event, [] if missing
            event_list = self.events.get(event, [])

            # Don't schedule if the event is repeating and already scheduled
            if repeat and event in self.events:
                LOG.debug(f'Repeating event {event} is already scheduled, '
                          f'discarding')
            else:
                LOG.debug(f"Scheduled event: {event} for time {sched_time}")
                # add received event and time
                event_list.append((sched_time, repeat, data, context))
                self.events[event] = event_list
                if sched_time < time.time():
                    LOG.warning(f"Added event is scheduled in the past and "
                                f"will be called immediately: {event}")

    def schedule_event_handler(self, message: Message):
        """
        Messagebus interface to the schedule_event method.
        Required data in the message envelope is
            event: event to emit
            time:  time to emit the event

        Optional data is
            repeat: repeat interval
            data:   data to send along with the event
        """
        event = message.data.get('event')
        sched_time = message.data.get('time')
        repeat = message.data.get('repeat')
        data = message.data.get('data')
        context = message.context
        if event and sched_time:
            self.schedule_event(event, sched_time, repeat, data, context)
        elif not event:
            LOG.error('Scheduled event name not provided')
        else:
            LOG.error('Scheduled event time not provided')

    def remove_event(self, event: str):
        """
        Remove an event from the list of scheduled events.

        Arguments:
            event (str): event identifier
        """
        with self.event_lock:
            if event in self.events:
                self.events.pop(event)

    def remove_event_handler(self, message: Message):
        """
        Messagebus interface to the remove_event method.
        """
        event = message.data.get('event')
        self.remove_event(event)

    def update_event(self, event: str, data: dict):
        """
        Change an existing event's data.

        This will only update the first call if multiple calls are registered
        to the same event identifier.

        Arguments:
            event (str): event identifier
            data (dict): new data
        """
        with self.event_lock:
            # if there is an active event with this name
            if len(self.events.get(event, [])) > 0:
                event_time, repeat, _, context = self.events[event][0]
                self.events[event][0] = (event_time, repeat, data, context)

    def update_event_handler(self, message: Message):
        """
        Messagebus interface to the update_event method.
        """
        event = message.data.get('event')
        data = message.data.get('data')
        self.update_event(event, data)

    def get_event_handler(self, message: Message):
        """
        Messagebus interface to get_event.

        Emits another event sending event status.
        """
        event_name = message.data.get("name")
        event = None
        with self.event_lock:
            if event_name in self.events:
                event = self.events[event_name]
        emitter_name = f'mycroft.event_status.callback.{event_name}'
        self.bus.emit(message.reply(emitter_name, data=event))

    def store(self):
        """
        Write current schedule to disk.
        """
        with self.event_lock:
            with open(self.schedule_file, 'w') as schedule_file:
                json.dump(self.events, schedule_file)

    def clear_repeating(self):
        """
        Remove repeating events from events dict.
        """
        with self.event_lock:
            for evt in self.events:
                self.events[evt] = [tup for tup in self.events[evt]
                                    if tup[1] is None]

    def clear_empty(self):
        """
        Remove empty event entries from events dict.
        """
        with self.event_lock:
            self.events = {k: self.events[k] for k in self.events
                           if self.events[k] != []}

    def shutdown(self):
        """
        Stop the running thread.
        """
        try:
            self._stopping.set()
            # Remove listeners
            self.bus.remove_all_listeners('mycroft.scheduler.schedule_event')
            self.bus.remove_all_listeners('mycroft.scheduler.remove_event')
            self.bus.remove_all_listeners('mycroft.scheduler.update_event')
            # Wait for thread to finish
            self.join(30)
            # Prune event list in preparation for saving
            self.clear_repeating()
            self.clear_empty()
            # Store all pending scheduled events
            self.store()
            self._running.clear()
        except Exception as e:
            self._running.clear()
            if not isinstance(e, OSError):
                self.store()
            raise e


class EventContainer(_EventContainer):
    def __init__(self, *args, **kwargs):
        log_deprecation("Import from `ovos_utils.events`", "0.1.0")
        _EventContainer.__init__(self, *args, **kwargs)


class EventSchedulerInterface(_SchedulerInterface):
    def __init__(self, *args, **kwargs):
        log_deprecation("Import from `ovos_utils.events`", "0.1.0")
        _SchedulerInterface.__init__(self, *args, **kwargs)



from websocket import create_connection

from ovos_config.config import Configuration
from ovos_bus_client.client import MessageBusClient
from ovos_bus_client.message import Message


def send(message_to_send: str, data_to_send: dict = None):
    """Send a single message over the websocket.

    Args:
        message_to_send (str): Message to send
        data_to_send (dict): data structure to go along with the
            message, defaults to empty dict.
    """
    data_to_send = data_to_send or {}

    # Calculate the standard Mycroft messagebus websocket address
    config = Configuration()
    config = config.get("websocket")
    url = MessageBusClient.build_url(
        config.get("host"),
        config.get("port"),
        config.get("route"),
        config.get("ssl")
    )

    # Send the provided message/data
    ws = create_connection(url)
    packet = Message(message_to_send, data_to_send).serialize()
    ws.send(packet)
    ws.close()



#!/usr/bin/env python3
# each method here is a console_script defined in setup.py
# each corresponds to a cli util
from ovos_bus_client import MessageBusClient, Message
from ovos_config import Configuration
import sys
import time


def ovos_speak():
    args_count = len(sys.argv)
    if args_count == 2:
        utt = sys.argv[1]
        lang = Configuration().get("lang", "en-us")
    elif args_count == 3:
        utt = sys.argv[1]
        lang = sys.argv[2]
    else:
        print("USAGE: ovos-speak {utterance} [lang]")
        raise SystemExit(2)
    client = MessageBusClient()
    client.run_in_thread()
    if not client.connected_event.is_set():
        client.connected_event.wait()
    client.emit(Message("speak", {"utterance": utt, "lang": lang}))
    time.sleep(0.5)  # avoids crash in c++ bus server
    client.close()


def ovos_say_to():
    args_count = len(sys.argv)
    if args_count == 2:
        utt = sys.argv[1]
        lang = Configuration().get("lang", "en-us")
    elif args_count == 3:
        utt = sys.argv[1]
        lang = sys.argv[2]
    else:
        print("USAGE: ovos-say-to {utterance} [lang]")
        raise SystemExit(2)
    client = MessageBusClient()
    client.run_in_thread()
    if not client.connected_event.is_set():
        client.connected_event.wait()
    client.emit(Message("recognizer_loop:utterance", {"utterances": [utt], "lang": lang}))
    time.sleep(0.5)  # avoids crash in c++ bus server
    client.close()


def ovos_listen():
    client = MessageBusClient()
    client.run_in_thread()
    if not client.connected_event.is_set():
        client.connected_event.wait()
    client.emit(Message("mycroft.mic.listen"))
    time.sleep(0.5)  # avoids crash in c++ bus server
    client.close()


def simple_cli():
    args_count = len(sys.argv)
    if args_count == 1:
        lang = Configuration().get("lang", "en-us")
    elif args_count == 2:
        lang = sys.argv[1]
    else:
        print("USAGE: ovos-simple-cli [lang]")
        return

    client = MessageBusClient()
    client.run_in_thread()
    if not client.connected_event.is_set():
        client.connected_event.wait()
    lang = lang or Configuration().get("lang", "en-us")

    from ovos_bus_client.session import SessionManager, Session
    sess = SessionManager.default_session

    while True:
        try:
            utt = input("Say:")
            if utt == ":exit":
                break
            client.emit(Message("recognizer_loop:utterance",
                                {"utterances": [utt], "lang": lang},
                                {"session": sess.serialize()}))
            time.sleep(0.5)  # avoids crash in c++ bus server
        except KeyboardInterrupt:
            break

    client.close()


if __name__ == "__main__":
    simple_cli()



import orjson
import time
import traceback
from os import getpid
from threading import Event, Thread
from typing import Union, Callable, Any, List, Optional
from uuid import uuid4

from ovos_utils.log import LOG, deprecated
from pyee import ExecutorEventEmitter
from websocket import (WebSocketApp,
                       WebSocketConnectionClosedException,
                       WebSocketException)

from ovos_bus_client.client.collector import MessageCollector
from ovos_bus_client.client.waiter import MessageWaiter
from ovos_bus_client.conf import load_message_bus_config, MessageBusClientConf, load_gui_message_bus_config
from ovos_bus_client.message import Message, CollectionMessage, GUIMessage
from ovos_bus_client.session import SessionManager, Session

try:
    from mycroft_bus_client import MessageBusClient as _MessageBusClientBase
except ImportError:
    # TODO - code in the wild does isinstance checks
    # this conditional subclassing should be removed ASAP, it is only here for the migration period
    # mycroft_bus_client is abandonware until further notice from MycroftAI

    class _MessageBusClientBase:
        pass


class MessageBusClient(_MessageBusClientBase):
    """The Mycroft Messagebus Client

    The Messagebus client connects to the Mycroft messagebus service
    and allows communication with the system. It has been extended to work
    like the pyee EventEmitter and tries to offer as much convenience as
    possible to the developer.
    """
    # minimize reading of the .conf
    _config_cache = None

    def __init__(self, host=None, port=None, route=None, ssl=None,
                 emitter=None, cache=False, session=None):
        config_overrides = dict(host=host, port=port, route=route, ssl=ssl)
        if cache and self._config_cache:
            config = self._config_cache
        else:
            config = load_message_bus_config(**config_overrides)
            if cache:
                MessageBusClient._config_cache = config

        self.config = MessageBusClientConf(config.host, config.port,
                                           config.route, config.ssl)
        self.emitter = emitter or ExecutorEventEmitter()
        self.client = self.create_client()
        self.retry = 5
        self.connected_event = Event()
        self.started_running = False
        self.wrapped_funcs = {}
        if session:
            SessionManager.update(session)
        else:
            session = SessionManager.default_session

        self.session_id = session.session_id
        self.on("ovos.session.update_default",
                self.on_default_session_update)

    @staticmethod
    def build_url(host: str, port: int, route: str, ssl: bool) -> str:
        """
        Build a websocket url.
        """
        return f"{'wss' if ssl else 'ws'}://{host}:{port}{route}"

    def create_client(self) -> WebSocketApp:
        """
        Setup websocket client.
        """
        url = self.build_url(ssl=self.config.ssl,
                             host=self.config.host,
                             port=self.config.port,
                             route=self.config.route)
        return WebSocketApp(url, on_open=self.on_open, on_close=self.on_close,
                            on_error=self.on_error, on_message=self.on_message)

    def on_open(self, *args):
        """
        Handle the "open" event from the websocket.
        A Basic message with the name "open" is forwarded to the emitter.
        """
        LOG.debug("Connected")
        self.connected_event.set()
        self.emitter.emit("open")
        # Restore reconnect timer to 5 seconds on sucessful connect
        self.retry = 5
        self.emit(Message("ovos.session.sync")) # request default session update

    def on_close(self, *args):
        """
        Handle the "close" event from the websocket.
        A Basic message with the name "close" is forwarded to the emitter.
        """
        self.emitter.emit("close")

    def on_error(self, *args):
        """
        On error start trying to reconnect to the websocket.
        """
        if len(args) == 1:
            error = args[0]
        else:
            error = args[1]
        if isinstance(error, WebSocketConnectionClosedException):
            LOG.warning('Could not send message because connection has closed')
        elif isinstance(error, ConnectionRefusedError):
            LOG.warning('Connection Refused. Is Messagebus Service running?')
        elif isinstance(error, ConnectionResetError):
            LOG.warning('Connection Reset. Did the Messagebus Service stop?')
        else:
            LOG.exception('=== %s ===', repr(error))
            try:
                self.emitter.emit('error', error)
            except Exception as e:
                LOG.exception(f'Failed to emit error event: {e}')

        try:
            if self.client.keep_running:
                self.client.close()
        except Exception as e:
            LOG.error(f'Exception closing websocket at {self.client.url}: {e}')

        LOG.warning("Message Bus Client "
                    "will reconnect in %.1f seconds.", self.retry)
        time.sleep(self.retry)
        self.retry = min(self.retry * 2, 60)
        try:
            self.emitter.emit('reconnecting')
            self.client = self.create_client()
            self.run_forever()
        except WebSocketException:
            pass

    def on_message(self, *args):
        """
        Handle an incoming websocket message
        @param args:
            message (str): serialized Message
        """
        if len(args) == 1:
            message = args[0]
        else:
            message = args[1]
        parsed_message = Message.deserialize(message)
        sess = Session.from_message(parsed_message)
        if sess.session_id != "default": # 'default' can only be updated by core
            SessionManager.update(sess)
        self.emitter.emit('message', message)
        self.emitter.emit(parsed_message.msg_type, parsed_message)

    def on_default_session_update(self, message):
        new_session = message.data["session_data"]
        sess = Session.deserialize(new_session)
        SessionManager.update(sess, make_default=True)
        LOG.debug("synced default_session")

    def emit(self, message: Message):
        """
        Send a message onto the message bus.

        This will both send the message to the local process using the
        event emitter and onto the Mycroft websocket for other processes.

        Args:
            message (Message): Message to send
        """
        if "session" not in message.context:
            sess = SessionManager.sessions.get(self.session_id) or \
                   Session(self.session_id)
            message.context["session"] = sess.serialize()

        if not self.connected_event.wait(10):
            if not self.started_running:
                raise ValueError('You must execute run_forever() '
                                 'before emitting messages')
            self.connected_event.wait()

        if hasattr(message, 'serialize'):
            msg = message.serialize()
        else:
            msg = orjson.dumps(message.__dict__).decode("utf-8")
        try:
            self.client.send(msg)
        except WebSocketConnectionClosedException:
            LOG.warning(f'Could not send {message.msg_type} message because connection '
                        'has been closed')
        except Exception as e:
            LOG.exception(f"failed to emit message {message.msg_type} with len {len(msg)}")

    def collect_responses(self, message: Message,
                          min_timeout: Union[int, float] = 0.2,
                          max_timeout: Union[int, float] = 3.0,
                          direct_return_func: Callable[[Message], Any] =
                          lambda msg: False) -> List[Message]:
        """
        Collect responses from multiple handlers.

        This sets up a collect-call (pun intended) expecting multiple handlers
        to respond.

        Args:
            message (Message): message to send
            min_timeout (int/float): Minimum time to wait for a response
            max_timeout (int/float): Maximum allowed time to wait for an answer
            direct_return_func (callable): Optional function for allowing an
                early return (not all registered handlers need to respond)

            Returns:
                (list) collected response messages.
        """
        collector = MessageCollector(self, message,
                                     min_timeout, max_timeout,
                                     direct_return_func)
        return collector.collect()

    def on_collect(self, event_name: str,
                   func: Callable[[CollectionMessage], Any],
                   timeout: Union[int, float] = 2):
        """
        Create a handler for a collect_responses call.

        This immeditely responds with an ack to register the handler with
        the caller, promising to return a response.

        The handler function then needs to send a response.

        Args:
            event_name (str): Message type to listen for.
            func (callable): function / method do be called for processing the
                             message.
            timeout (int/float): optional timeout of the handler
        """

        def wrapper(msg):
            collect_id = msg.context['__collect_id__']
            handler_id = str(uuid4())
            # Immediately respond that something is working on the issue
            acknowledge = msg.reply(msg.msg_type + '.handling',
                                    data={'query': collect_id,
                                          'handler': handler_id,
                                          'timeout': timeout})
            self.emit(acknowledge)
            func(CollectionMessage.from_message(msg, handler_id, collect_id))

        self.wrapped_funcs[func] = wrapper
        self.on(event_name, wrapper)

    def wait_for_message(self, message_type: str,
                         timeout: Union[int, float] = 3.0) -> Optional[Message]:
        """
        Wait for a message of a specific type.

        Arguments:
            message_type (str): the message type of the expected message
            timeout: seconds to wait before timeout, defaults to 3

        Returns:
            The received message or None if the response timed out
        """

        return MessageWaiter(self, message_type).wait(timeout)

    def wait_for_response(self, message: Message,
                          reply_type: Optional[Union[str, List[str]]] = None,
                          timeout: Union[float, int] = 3.0) -> \
            Optional[Message]:
        """
        Send a message and wait for a response.

        Arguments:
            message (Message): message to send
            reply_type (str | List[str]): the message type(s) of the expected reply.
                              Defaults to "<message.msg_type>.response".
            timeout: seconds to wait before timeout, defaults to 3

        Returns:
            The received message or None if the response timed out
        """
        message_type = None
        if isinstance(reply_type, list):
            message_type = reply_type
        elif isinstance(reply_type, str):
            message_type = [reply_type]
        elif reply_type is None:
            message_type = [message.msg_type + '.response']
        waiter = MessageWaiter(self, message_type)  # Setup response handler
        # Send message and wait for its response
        self.emit(message)
        return waiter.wait(timeout)

    def on(self, event_name: str, func: Callable[[Message], Any]):
        """Register callback with event emitter.

        Args:
            event_name (str): message type to map to the callback
            func (callable): callback function
        """
        self.emitter.on(event_name, func)

    def once(self, event_name: str, func: Callable[[Message], Any]):
        """Register callback with event emitter for a single call.

        Args:
            event_name (str): message type to map to the callback
            func (callable): callback function
        """
        self.emitter.once(event_name, func)

    def remove(self, event_name: str, func: Callable[[Message], Any]):
        """Remove registered event.

        Args:
            event_name (str): message type to map to the callback
            func (callable): callback function
        """
        if func in self.wrapped_funcs:
            self._remove_wrapped(event_name, func)
        else:
            self._remove_normal(event_name, func)

    def _remove_wrapped(self, event_name, external_func):
        """Remove a wrapped function."""

        wrapper = self.wrapped_funcs.pop(external_func)
        self._remove_normal(event_name, wrapper)

    def _remove_normal(self, event_name, func):
        try:
            if event_name not in self.emitter._events:
                LOG.debug("Not able to find '%s'", event_name)
            self.emitter.remove_listener(event_name, func)
        except (ValueError, KeyError):
            LOG.warning('Failed to remove event %s: %s',
                        event_name, str(func))
            for line in traceback.format_stack():
                LOG.warning(line.strip())

            if event_name not in self.emitter._events:
                LOG.debug("Not able to find '%s'", event_name)
            LOG.warning("Existing events: %s", repr(self.emitter._events))
            for evt in self.emitter._events:
                LOG.warning("   %s", repr(evt))
                LOG.warning("       %s", repr(self.emitter._events[evt]))
            if event_name in self.emitter._events:
                LOG.debug("Removing found '%s'", event_name)
            else:
                LOG.debug("Not able to find '%s'", event_name)
            LOG.warning('----- End dump -----')

    def remove_all_listeners(self, event_name: str):
        """
        Remove all listeners connected to event_name.

        Arguments:
            event_name: event from which to remove listeners
        """
        if event_name is None:
            raise ValueError
        self.emitter.remove_all_listeners(event_name)

    def run_forever(self):
        """
        Start the websocket handling.
        """
        self.started_running = True
        self.client.run_forever()

    def close(self):
        """
        Close the websocket connection.
        """
        self.client.close()
        self.connected_event.clear()

    def run_in_thread(self):
        """Launches the run_forever in a separate daemon thread."""
        t = Thread(target=self.run_forever)
        t.daemon = True
        t.start()
        return t


class GUIWebsocketClient(MessageBusClient):

    def __init__(self, host=None, port=None, route=None, ssl=None,
                 emitter=None, cache=False, client_name="ovos-gui-client"):
        self.gui_id = f"{client_name}_{getpid()}"
        config_overrides = dict(host=host, port=port, route=route, ssl=ssl)
        config = load_gui_message_bus_config(**config_overrides)
        super().__init__(host=config.host, port=config.port, route=config.route,
                         ssl=config.ssl, emitter=emitter, cache=cache)

    def emit(self, message: GUIMessage):
        """
        Send a message onto the message bus.

        This will both send the message to the local process using the
        event emitter and onto the Mycroft websocket for other processes.

        Args:
            message (GUIMessage): Message to send
        """

        if not self.connected_event.wait(10):
            if not self.started_running:
                raise ValueError('You must execute run_forever() '
                                 'before emitting messages')
            self.connected_event.wait()

        try:
            if hasattr(message, 'serialize'):
                self.client.send(message.serialize())
            else:
                self.client.send(orjson.dumps(message.__dict__).decode("utf-8"))
        except WebSocketConnectionClosedException:
            LOG.warning('Could not send %s message because connection '
                        'has been closed', message.msg_type)

    def on_open(self, *args):
        super().on_open(*args)
        self.emit(GUIMessage("mycroft.gui.connected",
                             gui_id=self.gui_id))

    def on_message(self, *args):
        """
        Handle an incoming websocket message
        @param args:
            message (str): serialized Message
        """
        if len(args) == 1:
            message = args[0]
        else:
            message = args[1]

        self.emitter.emit('message', message)

        parsed_message = GUIMessage.deserialize(message)
        self.emitter.emit(parsed_message.msg_type, parsed_message)


@deprecated("No direct replacement", "0.1.0")
def echo():
    """
    Echo function repeating all input from a user.
    """

    from ovos_bus_client.util import create_echo_function
    # TODO: Deprecate in 0.1.0
    message_bus_client = MessageBusClient()

    def repeat_utterance(message):
        message.msg_type = 'speak'
        message_bus_client.emit(message)

    message_bus_client.on('message', create_echo_function(None))
    message_bus_client.on('recognizer_loop:utterance', repeat_utterance)
    message_bus_client.run_forever()



# Copyright 2021 Mycroft AI Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""
Small utils and tools to use with the Messagebus.
"""

import logging
from typing import Callable, Optional, Union
from ovos_bus_client.message import Message
from ovos_utils.log import deprecated


@deprecated("No direct replacement", "0.1.0")
def create_echo_function(name: Optional[str]) -> \
        Callable[[Union[Message, str]], None]:
    """
    Standard logging mechanism for Mycroft processes.

    Arguments:
        name (str): Reference name of the process

    Returns:
        func: The echo function
    """
    # TODO: Deprecate in 0.1.0
    log = logging.getLogger(name)

    def echo(message: Union[Message, str]):
        try:
            if isinstance(message, str):
                msg = Message.deserialize(message)
            else:
                msg = message
            # do not log tokens from registration messages
            if msg.msg_type == "registration":
                msg.data["token"] = None
                message = msg.serialize()
        except Exception as exc:
            log.info(f"Error: {exc}", exc_info=True)

        # Listen for messages and echo them for logging
        log.info(f"BUS: {message}")
    return echo



import enum
import time
from threading import Lock, Event
from typing import Optional, List, Tuple, Union, Iterable, Dict
from uuid import uuid4

from ovos_config.config import Configuration
from ovos_config.locale import get_default_lang
from ovos_utils.log import LOG, log_deprecation

from ovos_bus_client.message import dig_for_message, Message


class UtteranceState(str, enum.Enum):
    INTENT = "intent"  # includes converse
    RESPONSE = "response"


class IntentContextManagerFrame:
    def __init__(self, entities: List[dict] = None, metadata: Dict = None):
        """
        Manages entities and context for a single frame of conversation.
        Provides simple equality querying.
        Attributes:
            entities(list): Entities that belong to ContextManagerFrame
            metadata(object): metadata to describe context belonging to ContextManagerFrame
        """
        self.entities = entities or []
        self.metadata = metadata or {}

    def serialize(self) -> dict:
        """
        Get a dict representation of this frame
        """
        return {"entities": self.entities,
                "metadata": self.metadata}

    @staticmethod
    def deserialize(data: Dict):
        """
        Build an IntentContextManagerFrame from serialized data
        @param data: serialized (dict) frame data
        @return: IntentContextManagerFrame for the specified data
        """
        return IntentContextManagerFrame(**data)

    def metadata_matches(self, query: Dict = None) -> bool:
        """
        Returns key matches to metadata
        Asserts that the contents of query exist within (logical subset of)
        metadata in this frame.
        Args:
            query(dict): metadata for matching
        Returns:
            bool:
                True: when key count in query is > 0 and all keys in query in
                    self.metadata
                False: if key count in query is <= 0 or any key in query not
                    found in self.metadata
        """
        query = query or {}
        result = len(query.keys()) > 0
        for key in query.keys():
            result = result and query[key] == self.metadata.get(key)

        return result

    def merge_context(self, tag: Dict, metadata: Dict):
        """
        merge into contextManagerFrame new entity and metadata.
        Appends tag as new entity and adds keys in metadata to keys in
        self.metadata.
        Args:
            tag(dict): entity to be added to self.entities
            metadata(dict): metadata contains keys to be added to self.metadata
        """
        self.entities.append(tag)
        for k, v in metadata.items():
            if k not in self.metadata:
                self.metadata[k] = v


class IntentContextManager:
    """
    Context Manager

    Use to track context throughout the course of a conversational session.
    How to manage a session's lifecycle is not captured here.
    """

    def __init__(self, timeout: int = None,
                 frame_stack: List[Tuple[IntentContextManagerFrame,
                 float]] = None,
                 greedy: bool = None, keywords: List[str] = None,
                 max_frames: int = None):

        config = Configuration().get('context', {})
        if timeout is None:
            timeout = config.get('timeout', 2) * 60  # minutes to seconds
        if greedy is None:
            greedy = config.get('greedy', False)
        if keywords is None:
            keywords = config.get('keywords', [])
        if max_frames is None:
            max_frames = config.get('max_frames', 3)

        self.frame_stack = frame_stack or []
        self.timeout = timeout
        self.context_keywords = keywords
        self.context_max_frames = max_frames
        self.context_greedy = greedy

    def serialize(self) -> dict:
        """
        Get a dict representation of this IntentContextManager
        """
        return {"timeout": self.timeout,
                "frame_stack": [(s.serialize(), t) for (s, t)
                                in self.frame_stack]}

    @staticmethod
    def deserialize(data: Dict):
        """
        Build an IntentContextManager from serialized data
        @param data: serialized (dict) data
        @return: IntentContextManager for the specified data
        """
        timeout = data.get("timeout", 2 * 60)
        framestack = [(IntentContextManagerFrame.deserialize(f), t)
                      for (f, t) in data.get("frame_stack", [])]
        return IntentContextManager(timeout, framestack)

    def update_context(self, entities: Dict):
        """
        Updates context with keyword from the intent.

        entity(dict): Format example...
                   {'data': 'Entity tag as <str>',
                    'key': 'entity proper name as <str>',
                    'confidence': <float>'
                   }

        Args:
            entities (list): Intent to scan for keywords
        """
        for context_entity in entities:
            if self.context_greedy:
                self.inject_context(context_entity)
            elif context_entity['data'][0][1] in self.context_keywords:
                self.inject_context(context_entity)

    def clear_context(self):
        """Remove all contexts."""
        self.frame_stack = []

    def remove_context(self, context_id: str):
        """Remove a specific context entry.

        Args:
            context_id (str): context entry to remove
        """
        self.frame_stack = [(f, t) for (f, t) in self.frame_stack
                            if context_id in f.entities[0].get('data', [])]

    def inject_context(self, entity: Dict, metadata: Dict = None):
        """
        Add context to the first frame in the stack. If no frame metadata
        doesn't match the passed metadata then a new one is inserted.
        Args:
            entity(dict): Format example...
                       {'data': 'Entity tag as <str>',
                        'key': 'entity proper name as <str>',
                        'confidence': <float>'
                       }
            metadata(dict): arbitrary metadata about entity injected
        """
        metadata = metadata or {}
        try:
            if self.frame_stack:
                top_frame = self.frame_stack[0]
            else:
                top_frame = None
            if top_frame and top_frame[0].metadata_matches(metadata):
                top_frame[0].merge_context(entity, metadata)
            else:
                frame = IntentContextManagerFrame(entities=[entity],
                                                  metadata=metadata.copy())
                self.frame_stack.insert(0, (frame, time.time()))
        except (IndexError, KeyError):
            pass

    @staticmethod
    def _strip_result(context_features: Iterable):
        """Keep only the latest instance of each keyword.

        Arguments
            context_features (iterable): context features to check.
        """
        stripped = []
        processed = []
        for feature in context_features:
            keyword = feature['data'][0][1]
            if keyword not in processed:
                stripped.append(feature)
                processed.append(keyword)
        return stripped

    def get_context(self, max_frames: int = None,
                    missing_entities: List[str] = None):
        """
        Constructs a list of entities from the context.

        Args:
            max_frames(int): maximum number of frames to look back
            missing_entities(list of str): a list or set of tag names,
            as strings

        Returns:
            list: a list of entities
        """
        missing_entities = missing_entities or []

        relevant_frames = [frame[0] for frame in self.frame_stack if
                           time.time() - frame[1] < self.timeout]
        if not max_frames or max_frames > len(relevant_frames):
            max_frames = len(relevant_frames)

        missing_entities = list(missing_entities)
        context = []
        last = ''
        depth = 0
        entity = {}
        for i in range(max_frames):
            frame_entities = [entity.copy() for entity in
                              relevant_frames[i].entities]
            for entity in frame_entities:
                entity['confidence'] = entity.get('confidence', 1.0) \
                                       / (2.0 + depth)
            context += frame_entities

            # Update depth
            if entity['origin'] != last or entity['origin'] == '':
                depth += 1
            last = entity['origin']

        result = []
        if missing_entities:
            for entity in context:
                if entity.get('data') in missing_entities:
                    result.append(entity)
                    # NOTE: this implies that we will only ever get one
                    # of an entity kind from context, unless specified
                    # multiple times in missing_entities. Cannot get
                    # an arbitrary number of an entity kind.
                    missing_entities.remove(entity.get('data'))
        else:
            result = context

        # Only use the latest  keyword
        return self._strip_result(result)


class Session:
    def __init__(self, session_id: str = None,
                 expiration_seconds: int = None,
                 active_skills: List[List[Union[str, float]]] = None,
                 utterance_states: Dict = None,
                 lang: str = None,
                 context: IntentContextManager = None,
                 site_id: str = "unknown",
                 pipeline: List[str] = None,
                 stt_prefs: Dict = None,
                 tts_prefs: Dict = None,
                 location_prefs: Dict = None,
                 system_unit: str = None,
                 time_format: str = None,
                 date_format: str = None,
                 is_speaking: bool = False,
                 is_recording: bool = False,
                 blacklisted_intents: Optional[List[str]] = None,
                 blacklisted_skills: Optional[List[str]] = None):
        """
        Construct a session identifier
        @param session_id: string UUID for the session
        @param expiration_seconds: TTL for session (-1 for no expiration)
        @param active_skills: List of list skill_id, last reference
        @param utterance_states: dict of skill_id to UtteranceState
        @param lang: language associated with this Session
        @param context: IntentContextManager for this Session
        """
        if tts_prefs:
            log_deprecation("'tts_prefs' kwarg has been deprecated! value will be ignored", "0.1.0")
        if stt_prefs:
            log_deprecation("'stt_prefs' kwarg has been deprecated! value will be ignored", "0.1.0")
        self.session_id = session_id or str(uuid4())
        self.blacklisted_skills = (blacklisted_skills or
                                   Configuration().get("skills", {}).get("blacklisted_skills", []))
        self.blacklisted_intents = (blacklisted_intents or
                                    Configuration().get("intents", {}).get("blacklisted_intents", []))
        self.lang = lang or get_default_lang()
        self.system_unit = system_unit or Configuration().get("system_unit", "metric")
        self.date_format = date_format or Configuration().get("date_format", "DMY")
        self.time_format = time_format or Configuration().get("time_format", "full")

        self.is_recording = is_recording
        self.is_speaking = is_speaking
        self.site_id = site_id or Configuration().get("site_id") or "unknown"  # indoors placement info

        self.active_skills = active_skills or []  # [skill_id , timestamp]# (Message , timestamp)
        self.utterance_states = utterance_states or {}  # {skill_id: UtteranceState}

        self.touch_time = int(time.time())
        self.expiration_seconds = expiration_seconds or \
                                  Configuration().get('session', {}).get("ttl", -1)
        self.pipeline = pipeline or Configuration().get('intents', {}).get("pipeline") or [
            "stop_high",
            "converse",
            "padatious_high",
            "adapt_high",
            "fallback_high",
            "stop_medium",
            "padatious_medium",
            "adapt_medium",
            "adapt_low",
            "common_qa",
            "fallback_medium",
            "fallback_low"
        ]
        self.context = context or IntentContextManager()

        self.location_preferences = location_prefs or Configuration().get("location", {})

    @property
    def active(self) -> bool:
        """
        Return true if any skills attached to this session are active.
        NOTE: skills without converse implemented never get added here unless
        using get_response
        """
        return len(self.active_skills) > 0

    def touch(self):
        """
        update the touch_time on the session
        """
        self.touch_time = int(time.time())
        SessionManager.update(self)

    def expired(self) -> bool:
        """
        Return True if the session has expired
        """
        if self.expiration_seconds < 0:
            return False
        return int(time.time()) - self.touch_time > self.expiration_seconds

    def __str__(self):
        return "{%s,%d}" % (str(self.session_id), self.touch_time)

    def enable_response_mode(self, skill_id: str):
        """
        Mark a skill as expecting a response
        @param skill_id: ID of skill expecting a response
        """
        self.utterance_states[skill_id] = UtteranceState.RESPONSE.value
        self.touch()

    def disable_response_mode(self, skill_id: str):
        """
        Mark a skill as not expecting a response (handling intents normally)
        @param skill_id: ID of skill expecting a response
        """
        self.utterance_states[skill_id] = UtteranceState.INTENT.value
        self.touch()

    def activate_skill(self, skill_id: str):
        """
        Add a skill to the front of the active_skills list
        @param skill_id: ID of skill to activate
        """
        # remove it from active list
        self.deactivate_skill(skill_id)
        # add skill with timestamp to start of active list
        self.active_skills.insert(0, [skill_id, time.time()])
        self.touch()

    def deactivate_skill(self, skill_id: str):
        """
        Remove a skill from the active_skills list
        @param skill_id: ID of skill to deactivate
        """
        active_ids = [s[0] for s in self.active_skills]
        if skill_id in active_ids:
            idx = active_ids.index(skill_id)
            self.active_skills.pop(idx)
        self.touch()

    def is_active(self, skill_id: str) -> bool:
        """
        Check if a skill is active
        @param skill_id: ID of skill to check
        @return: True if the requested skill is active
        """
        active_ids = [s[0] for s in self.active_skills]
        return skill_id in active_ids

    def clear(self):
        """
        Clear active_skills
        """
        self.active_skills = []
        self.touch()

    def serialize(self) -> dict:
        """
        Get a json-serializable dict representation of this session
        """
        # safe for json dumping
        return {
            "active_skills": self.active_skills,
            "utterance_states": self.utterance_states,
            "session_id": self.session_id,
            "lang": self.lang,
            "context": self.context.serialize(),
            "site_id": self.site_id,
            "pipeline": self.pipeline,
            "location": self.location_preferences,
            "system_unit": self.system_unit,
            "time_format": self.time_format,
            "date_format": self.date_format,
            "is_speaking": self.is_speaking,
            "is_recording": self.is_recording,
            "blacklisted_skills": self.blacklisted_skills,
            "blacklisted_intents": self.blacklisted_intents
        }

    def update_history(self, message: Message = None):
        """
        Add a message to history and then prune history
        @param message: Message to append to history
        """
        LOG.warning("update_history has been deprecated, "
                    "session no longer has a message history")

    @staticmethod
    def deserialize(data: Dict):
        """
        Build a Session object from dict data
        @param data: dict serialized Session object
        @return: Session representation of data
        """
        uid = data.get("session_id")
        active = data.get("active_skills") or []
        states = data.get("utterance_states") or {}
        lang = data.get("lang")
        context = IntentContextManager.deserialize(data.get("context", {}))
        site_id = data.get("site_id", "unknown")
        pipeline = data.get("pipeline", [])
        location = data.get("location", {})
        system_unit = data.get("system_unit")
        date_format = data.get("date_format")
        time_format = data.get("time_format")
        is_recording = data.get("is_recording", False)
        is_speaking = data.get("is_speaking", False)
        blacklisted_skills = data.get("blacklisted_skills", [])
        blacklisted_intents = data.get("blacklisted_intents", [])
        return Session(uid,
                       active_skills=active,
                       utterance_states=states,
                       lang=lang,
                       context=context,
                       pipeline=pipeline,
                       site_id=site_id,
                       location_prefs=location,
                       system_unit=system_unit,
                       date_format=date_format,
                       time_format=time_format,
                       is_recording=is_recording,
                       is_speaking=is_speaking,
                       blacklisted_intents=blacklisted_intents,
                       blacklisted_skills=blacklisted_skills)

    @staticmethod
    def from_message(message: Message = None):
        """
        Get a Session for the given message. If no session in message context,
        SessionManager.default_session is returned.
        If SessionManager.default_session is None, a default session is created
        @param message: Message to get session for
        @return: Session object
        """
        message = message or dig_for_message()
        if message and "session" in message.context:
            lang = message.context.get("lang") or \
                   message.data.get("lang")
            sess = message.context["session"]
            if "lang" not in sess:
                sess["lang"] = lang
            sess = Session.deserialize(sess)
        else:
            if message:
                LOG.warning(f"No session context in message:{message.msg_type}")
                LOG.debug(f"Update ovos-bus-client or add `session` to "
                          f"`message.context` where emitted. "
                          f"context={message.context}")
            else:
                LOG.warning(f"No message found, using default session")
            # new session
            sess = SessionManager.default_session
        if sess and sess.expired():
            LOG.debug(f"unexpiring session {sess.session_id}")
        return sess


class SessionManager:
    """ Keeps track of the current active session. """
    default_session: Session = Session("default")
    __lock = Lock()
    sessions = {"default": default_session}
    bus = None

    @classmethod
    def sync(cls, message=None):
        if cls.bus:
            message = message or Message("ovos.session.sync")
            cls.bus.emit(message.reply("ovos.session.update_default",
                                       {"session_data": cls.default_session.serialize()}))

    @classmethod
    def connect_to_bus(cls, bus):
        cls.bus = bus
        cls.bus.on("recognizer_loop:record_begin", cls.handle_recording_start)
        cls.bus.on("recognizer_loop:record_end", cls.handle_recording_end)
        cls.bus.on("recognizer_loop:audio_output_start", cls.handle_audio_output_start)
        cls.bus.on("recognizer_loop:audio_output_end", cls.handle_audio_output_end)
        cls.bus.on("ovos.session.sync", cls.handle_default_session_request)
        cls.sync()

    @staticmethod
    def prune_sessions():
        """
        Discard any expired sessions
        """
        # TODO: Consider when to prune sessions; an event or callback scheduled
        #   on `touch`, periodically scheduled event, or triggered on some
        #   interaction with `SessionManager` (ideally threaded to not slow
        #   down references)
        SessionManager.sessions = {sid: s for sid, s in
                                   SessionManager.sessions.items()
                                   if not s.expired}

    @staticmethod
    def reset_default_session() -> Session:
        """
        Define and return a new default_session
        """
        with SessionManager.__lock:
            sess = Session("default")
            LOG.info(f"Default Session reset")
            SessionManager.default_session = SessionManager.sessions["default"] = sess
            SessionManager.sync()
        return SessionManager.default_session

    @staticmethod
    def update(sess: Session, make_default: bool = False):
        """
        Update the last_touch timestamp on the current session
        @param sess: Session to update
        @param make_default: if true, set default_session to sess
        """
        if not sess:
            raise ValueError(f"Expected Session and got None")

        if make_default:
            sess.session_id = "default"
            # this log is dangerous, session may contain things like passwords and access keys
            # this comment is here to avoid reintroducing it by accident
            # LOG.debug(f"replacing default session with: {sess.serialize()}") # DO NOT re-enable in production

        if sess.session_id == "default":
            SessionManager.default_session = sess
        SessionManager.sessions[sess.session_id] = sess

    @staticmethod
    def get(message: Optional[Message] = None) -> Session:
        """
        Get the active session for a given Message

        @param message: Message to get session for
        @return: Session from message or default_session
        """
        sess = SessionManager.default_session
        message = message or dig_for_message()

        # A message exists, get a real session
        if message:
            msg_sess = Session.from_message(message)
            if msg_sess:
                if msg_sess.session_id != "default":  # reserved namespace for ovos-core
                    SessionManager.sessions[msg_sess.session_id] = msg_sess
                    return msg_sess
            else:
                LOG.debug(f"No session from message, use default session")
        else:
            LOG.debug(f"No message, use default session")

        return sess

    @staticmethod
    def touch(message: Message = None):
        """
        Update the last_touch timestamp on the current session

        @param message: Message to get Session for to update
        """
        sess = SessionManager.get(message)
        sess.touch()

    ##############################
    # util methods for skill consumption
    @classmethod
    def is_speaking(cls, session: Session = None):
        session = session or SessionManager.get()
        return session.is_speaking

    @classmethod
    def wait_while_speaking(cls, timeout=15, session: Session = None):
        """ wait until audio service reports end of audio output """
        if isinstance(timeout, bool):
            LOG.warning(f"expected timeout in seconds, got boolean '{timeout}', "
                        f"defaulting to 15 seconds")
            timeout = 15

        if not cls.bus:
            LOG.error("SessionManager not connected to bus, can not monitor speech state")
            return

        session = session or SessionManager.get()
        if not cls.is_speaking(session):
            LOG.warning(f"can't 'wait_while_speaking' because "
                        f"session '{session.session_id}' is not currently speaking")
            return

        # wait until end of speech
        LOG.debug(f"waiting for session '{session.session_id}' audio output to end with timeout: {timeout}")
        event = Event()
        sessid = session.session_id

        def handle_output_end(msg):
            nonlocal sessid, event
            sess = SessionManager.get(msg)
            if sessid == sess.session_id:
                LOG.debug(f"session: {sessid} audio output ended")
                event.set()

        cls.bus.on("recognizer_loop:audio_output_end", handle_output_end)
        event.wait(timeout=timeout)
        if not event.is_set():
            LOG.warning("waiting for audio output end timed out! not waiting anymore")
        cls.bus.remove("recognizer_loop:audio_output_end", handle_output_end)

    @classmethod
    def is_recording(cls, session: Session = None):
        session = session or SessionManager.get()
        return session.is_recording

    @classmethod
    def wait_while_recording(cls, timeout=45, session: Session = None):
        """ wait until listener service reports end of recording"""
        if not cls.bus:
            LOG.error("SessionManager not connected to bus, can not monitor recording state")
            return

        session = session or SessionManager.get()
        if not cls.is_recording(session):
            return

        # wait until end of recording
        event = Event()
        sessid = session.session_id

        def handle_rec_end(msg):
            nonlocal sessid, event
            sess = SessionManager.get(msg)
            if sessid == sess.session_id:
                event.set()

        cls.bus.on("recognizer_loop:record_end", handle_rec_end)
        event.wait(timeout=timeout)
        cls.bus.remove("recognizer_loop:record_end", handle_rec_end)

    ###############################
    # State tracking events
    @classmethod
    def handle_recording_start(cls, message):
        """track when a session is recording audio"""
        sess = cls.get(message)
        sess.is_recording = True
        cls.update(sess)

    @classmethod
    def handle_recording_end(cls, message):
        """track when a session stops recording audio"""
        sess = cls.get(message)
        sess.is_recording = False
        cls.update(sess)

    @classmethod
    def handle_audio_output_start(cls, message):
        """track when a session is outputting audio"""
        sess = cls.get(message)
        sess.is_speaking = True
        cls.update(sess)

    @classmethod
    def handle_audio_output_end(cls, message):
        """track when a session stops outputting audio"""
        sess = cls.get(message)
        sess.is_speaking = False
        cls.update(sess)

    @classmethod
    def handle_default_session_request(cls, message=None):
        cls.sync(message)



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from threading import Event
from typing import List, Union

try:
    from mycroft_bus_client.client.waiter import MessageWaiter as _MessageWaiterBase
except ImportError:
    # TODO - code in the wild does isinstance checks
    # this conditional subclassing should be removed ASAP, it is only here for the migration period
    # mycroft_bus_client is abandonware until further notice from MycroftAI

    class _MessageWaiterBase:
        pass


class MessageWaiter:
    """Wait for a single message.

    Encapsulate the wait for a message logic separating the setup from
    the actual waiting act so the waiting can be setup, actions can be
    performed and _then_ the message can be waited for.

    Arguments:
        bus: Bus to check for messages on
        message_type: message type(s) to wait for
    """
    def __init__(self, bus, message_type: Union[str, List[str]]):
        self.bus = bus
        if not isinstance(message_type, list):
            message_type = [message_type]
        self.msg_type = message_type
        self.received_msg = None
        # Setup response handler
        self.response_event = Event()
        for msg in self.msg_type:
            self.bus.once(msg, self._handler)

    def _handler(self, message):
        """Receive response data."""
        self.received_msg = message
        self.response_event.set()

    def wait(self, timeout=3.0):
        """Wait for message.

        Arguments:
            timeout (int or float): seconds to wait for message

        Returns:
            Message or None
        """
        self.response_event.wait(timeout)
        if not self.response_event.is_set():
            # Clean up the event handler
            try:
                for msg in self.msg_type:
                    self.bus.remove(msg, self._handler)
            except (ValueError, KeyError):
                # ValueError occurs on pyee 5.0.1 removing handlers
                # registered with once.
                # KeyError may theoretically occur if the event occurs as
                # the handler is removed
                pass
        return self.received_msg



# Copyright 2017 Mycroft AI Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from ovos_bus_client.client.client import MessageBusClient, GUIWebsocketClient
from ovos_bus_client.message import Message, GUIMessage
from ovos_bus_client.send_func import send
from ovos_bus_client.session import Session, SessionManager, UtteranceState
from ovos_bus_client.conf import client_from_config

"""
Mycroft Messagebus Client.

A client for using the Mycroft message bus.

The messagebus client allows the developer to register handlers for
specific message types and emitting messages to other services and
clients connected to the bus.
"""

__all__ = [
    MessageBusClient,
    GUIWebsocketClient,
    GUIMessage,
    Message,
    send,
    client_from_config
]
